#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
import parlai.utils.logging as logging
from typing import Dict, List, Optional
from collections import defaultdict
import copy
import json
import os
import random
import re
import torch

from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from nltk.stem.wordnet import WordNetLemmatizer
import nltk

import spacy
from checklist.perturb import Perturb
from parlai.tasks.reasoning.base import t_REASON_PROCESSED_RESULT, AbstractReason


class StepPerturbation:
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        return parser

    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[List[str]]]] = None):
        self.random = random.Random(42)
        self.cache = cache

    def perturb(self, example_dict: Dict) -> Dict:
        raise RuntimeError("Child class must implement")

    def batch_perturb(self, examples: List[Dict]) -> List[Dict]:
        """
        Override to parallelize expensive parts of perturb operation.
        """
        return [self.perturb(e) for e in examples]


class ShuffleSteps(StepPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) < 2:
            raise RuntimeError(
                "ShuffleSteps requires at least 2 steps.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        shuffled = copy.deepcopy(steps)
        # make sure steps are different from original
        while "".join([str(s) for s in shuffled]) == "".join([str(s) for s in steps]):
            # don't use shuffle cause that'll change the original + this a touch faster than reallocating
            shuffled = self.random.sample(steps, len(steps))
        example_dict['steps'] = shuffled
        return example_dict


class DuplicateOneStep(StepPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) == 0:
            raise RuntimeError(
                "DuplicateOneStep requires at least 1 step. Make sure `get_data_for_fold()`"
                + "appropriately filters for these turns when used with this perturbation"
            )
        to_dupe = self.random.choice(range(len(steps)))
        result = copy.deepcopy(steps)
        result[to_dupe:to_dupe] = [steps[to_dupe]]
        example_dict['steps'] = result
        return example_dict


class RemoveOneStep(StepPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) == 0:
            raise RuntimeError(
                "RemoveOneStep requires at least 1 step. Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation. "
            )
        to_dupe = self.random.choice(range(len(steps)))
        result = copy.deepcopy(steps)
        del result[to_dupe]
        example_dict['steps'] = result
        return example_dict


class SwapOneStep(StepPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) < 2:
            raise RuntimeError(
                "SwapOneStep requires at least 2 steps.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        idx_1 = idx_2 = 0
        result = copy.deepcopy(steps)
        while str(result[idx_1]) == str(result[idx_2]):
            to_dupe = self.random.sample(range(len(steps)), 2)
            idx_1, idx_2 = to_dupe[0], to_dupe[1]
        result[idx_1], result[idx_2] = result[idx_2], result[idx_1]
        example_dict['steps'] = result
        return example_dict


class ExtrinsicHallucinatedStep(StepPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if 'extrinsic_step' not in example_dict:
            raise RuntimeError(
                "ExtrinsicHallucinatedStep requires 'extrinsic_step' key.  Make sure `get_data_for_fold`()`"
                + "appropriately has this key in the yielded message to use this perturbation."
            )
        extrinsic_step = example_dict['extrinsic_step']
        if len(steps) == 0:
            raise RuntimeError("ExtrinsicHallucinatedStep requires at least 1 step")
        to_dupe = self.random.choice(range(len(steps)))
        result = copy.deepcopy(steps)
        result.insert(to_dupe, extrinsic_step)
        example_dict['steps'] = result
        return example_dict


class ParaphraseSteps(StepPerturbation):
    """
    Paraphrasing steps using model from https://huggingface.co/Vamsi/T5_Paraphrase_Paws.
    """

    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[List[str]]]] = None):
        super().__init__(opt, cache)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.tokenizer = AutoTokenizer.from_pretrained("Vamsi/T5_Paraphrase_Paws")
        self.model = AutoModelForSeq2SeqLM.from_pretrained(
            "Vamsi/T5_Paraphrase_Paws"
        ).to(self.device)
        self.model.eval()
        self.model.parallelize()

        self.cache = defaultdict(set)
        if cache is not None:
            self._populate_cache(cache)
        self.fillers = ['Well, ', 'Well we know ', 'Hm... ', 'Okay ']

    def _populate_cache(self, cache: Dict[str, List[List[str]]]) -> None:
        def _correct_type(value):
            if isinstance(value, str):
                return tuple(self._separate_sentences(value))
            if isinstance(value, list):
                return tuple(value)
            raise NotImplementedError(f'Unexpected value type: {value} ({type(value)})')

        for k, vs in cache.items():
            self.cache[k] = set(_correct_type(v) for v in vs)

    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        joined_steps = self._joined_steps(steps)

        if joined_steps not in self.cache:
            if self._reasoning_steps_contain_math(steps):
                # don't call paraphrase model or _separate_sentences
                paraphrased_steps = self._paraphrase_mathy_steps(steps)
                # add to cache
                self.cache[joined_steps].add(tuple(paraphrased_steps))
            else:
                # run paraphrase model
                result = self._batch_run_paraphrase_model([joined_steps])[0]

                # update cache
                if result == joined_steps:
                    logging.warn(
                        "ParaphraseSteps failed for the reasoning chain "
                        + joined_steps
                        + ". Make sure get_data_for_fold()` appropriately filters for these turns"
                        + "when used with this perturbation."
                    )
                    result = self._hard_coded_paraphrase(result)
                self.cache[joined_steps].add(tuple(self._separate_sentences(result)))

        example_dict['steps'] = self.random.choice(list(self.cache[joined_steps]))
        return example_dict

    def batch_perturb(self, examples: List[Dict]) -> List[Dict]:
        steps = [ex['steps'] for ex in examples]
        joined_steps = [(self._joined_steps(s), s) for s in steps]
        joined_steps_not_in_cache = [
            (js, os) for js, os in joined_steps if js not in self.cache
        ]

        if len(joined_steps_not_in_cache) > 0:
            mathy_steps = []
            texty_steps = []
            for js, s in joined_steps_not_in_cache:
                if self._reasoning_steps_contain_math(s):
                    mathy_steps.append((js, s))
                else:
                    texty_steps.append(js)

            for joined, s in mathy_steps:
                # don't call paraphrase model or _separate_sentences
                # add to cache
                self.cache[joined].add(tuple(self._paraphrase_mathy_steps(s)))

            if len(texty_steps) > 0:
                # run paraphrase model
                result = self._batch_run_paraphrase_model(texty_steps)

                # update cache
                for original, perturbed in zip(texty_steps, result):
                    if original == perturbed:
                        perturbed = self._hard_coded_paraphrase(perturbed)
                    self.cache[original].add(tuple(self._separate_sentences(perturbed)))

        # update examples from cache
        for i, (original, _) in enumerate(joined_steps):
            examples[i]['steps'] = self.random.choice(list(self.cache[original]))
        return examples

    def _joined_steps(self, steps: List[str]) -> str:
        if len(steps) == 0:
            raise RuntimeError(
                "ParaphraseSteps requires at least 1 step.  Make sure `get_data_for_fold()`"
                "appropriately filters for these turns when used with this perturbation."
            )
        return " ".join(str(step) for step in steps)

    def _reasoning_steps_contain_math(self, steps: List[str]) -> bool:
        for step in steps:
            if isinstance(step, MWPStep):
                return True
            # TODO: if step has math symbols ( +,-,=,*,/,^,@ ) and low concentration of letters
        return False

    def _paraphrase_mathy_steps(self, steps: List[str]) -> List[str]:
        # add fillers to first step
        steps[0] = self._hard_coded_paraphrase(str(steps[0]))
        # TODO: add fillers to other steps?
        # TODO: rearrange operations into equivalent statements
        return steps

    def _batch_run_paraphrase_model(self, texts: List[str]) -> List[str]:
        encoding = self.tokenizer.batch_encode_plus(
            [f"paraphrase: {s} </s>" for s in texts],
            pad_to_max_length=True,
            return_tensors="pt",
        )
        outputs = self.model.generate(
            input_ids=encoding["input_ids"].to(self.device),
            attention_mask=encoding["attention_mask"].to(self.device),
            max_length=256,
            do_sample=True,
            top_p=0.9,
            early_stopping=True,
        )
        result = self.tokenizer.batch_decode(
            outputs, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        return result

    def _hard_coded_paraphrase(self, chain_of_thought: str) -> str:
        if chain_of_thought.startswith('The'):
            chain_of_thought = 'the' + chain_of_thought[3:]
        return self.random.choice(self.fillers) + chain_of_thought

    def _separate_sentences(self, text: str) -> List[str]:
        text = text.lstrip(
            ' .!?'
        )  # nltk sentence tokenizer breaks when string starts with punctuation
        sents = nltk.tokenize.sent_tokenize(text)
        return [s for s in sents if len(s) > 1]  # remove standalone punctuation


class GrammaticalErrorStep(StepPerturbation):
    """
    Lemmatizes one step in example. - Changes the tense of the verbs in one step to infinitive

    Dependencies:
    import nltk
    nltk.download('wordnet')
    nltk.download('averaged_perceptron_tagger')


    and make sure wordnet gets unzipped in appropriate location
    """

    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[List[str]]]] = None):
        super().__init__(opt, cache)
        self.lemmatizer = WordNetLemmatizer()

    def lemmatize_step(self, words):
        lemmatized_output = ' '.join([self.lemmatizer.lemmatize(w, 'v') for w in words])
        # remove extraneous spaces after joining strings back
        clean_lemmatized_output = re.sub(
            r'\s([?.!"](?:\s|$))', r'\1', lemmatized_output
        )
        return clean_lemmatized_output

    def drop_verb(self, words):
        tags = nltk.pos_tag(words)
        verb_indices = []
        for i, tag in enumerate(tags):
            # keep track of indices of verbs
            if 'V' in tag[1]:
                verb_indices.append(i)
        if not verb_indices:
            return ""
        verb_to_drop = self.random.choice(range(len(verb_indices)))
        del words[verb_indices[verb_to_drop]]
        result = ' '.join(words)
        clean_result = re.sub(r'\s([?.!"](?:\s|$))', r'\1', result)
        return clean_result

    def swap_words(self, tokenized_step):
        tags = nltk.pos_tag(tokenized_step)
        word_indices = []
        for i, tag in enumerate(tags):
            # keep track of indices of words and not punctuation
            # only swaps words and leaves punctuation in place
            if tag[1].isalpha():
                word_indices.append(i)
        if len(word_indices) < 2:
            return ""
        to_dupe = self.random.sample(range(len(word_indices)), 2)
        idx_1, idx_2 = to_dupe[0], to_dupe[1]
        (tokenized_step[word_indices[idx_1]], tokenized_step[word_indices[idx_2]],) = (
            tokenized_step[word_indices[idx_2]],
            tokenized_step[word_indices[idx_1]],
        )
        result = ' '.join(tokenized_step)
        clean_result = re.sub(r'\s([?.!"](?:\s|$))', r'\1', result)
        return clean_result

    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) == 0:
            raise RuntimeError("NegateSteps requires at least 1 step")
        result = copy.deepcopy(steps)
        grammatical_error_steps = []

        for i, step in enumerate(steps):
            try:
                tok_step = nltk.word_tokenize(str(step))
            except IndexError:
                print(
                    f"WARNING: could not tokenize step {str(step)}. Proceeding to next chain."
                )
                return str(step)
            # perform all possible grammatical errors on each step, then randomly choose 1
            lemmatized_step = self.lemmatize_step(tok_step)
            if tok_step != lemmatized_step:
                grammatical_error_steps.append((i, lemmatized_step))
            dropped_verb_step = self.drop_verb(tok_step)
            if dropped_verb_step != "" and tok_step != dropped_verb_step:
                grammatical_error_steps.append((i, dropped_verb_step))
            swapped_word_step = self.swap_words(tok_step)
            if swapped_word_step != "" and tok_step != swapped_word_step:
                grammatical_error_steps.append((i, swapped_word_step))

        if not grammatical_error_steps:
            raise RuntimeError(
                "GrammaticalErrorStep failed for the reasoning chain "
                + " ".join(str(step) for step in steps)
                + ". Make sure get_data_for_fold()` appropriately filters for these turns when used with this perturbation."
            )

        to_dupe = self.random.choice(range(len(grammatical_error_steps)))
        result[grammatical_error_steps[to_dupe][0]] = grammatical_error_steps[to_dupe][
            1
        ]
        example_dict['steps'] = result
        return example_dict


class NegateStep(StepPerturbation):
    """
    Negates one step in example.

    Dependencies:
    pip3 install checklist
    python -m spacy download en_core_web_sm
    """

    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[List[str]]]] = None):
        super().__init__(opt, cache)
        self.nlp = spacy.load('en_core_web_sm')

    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        steps = example_dict['steps']
        if len(steps) == 0:
            raise RuntimeError(
                "NegateSteps requires at least 1 step. Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        result = copy.deepcopy(steps)
        negated_steps = []
        for i, step_to_dupe in enumerate(result):
            pdata = list(self.nlp.pipe([str(step_to_dupe)]))
            try:
                ret = Perturb.perturb(pdata, Perturb.add_negation)
                if ret.data:
                    negated_steps.append((i, ret.data[0][1]))
            # Perturb.add_negation is noted as still experimental in documentation - occasionally
            # fails on certain cases with a Runtime or Index Error
            except RuntimeError:
                pass
            except IndexError:
                pass
        if not negated_steps:
            raise RuntimeError(
                "NegateStep failed for the reasoning chain "
                + " ".join(str(step) for step in steps)
                + ". Make sure get_data_for_fold()` appropriately filters for these turns"
                + "when used with this perturbation."
            )
        to_dupe = self.random.choice(range(len(negated_steps)))
        result[negated_steps[to_dupe][0]] = negated_steps[to_dupe][1]
        example_dict['steps'] = result
        return example_dict


class SemanticChangeStep(StepPerturbation):
    """
    Change the semantics of one step in example by replacing entities.

    Dependencies:
    python -m spacy download en_core_web_sm
    """

    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[List[str]]]] = None):
        super().__init__(opt, cache)
        pattern = 'NP: {<DT>?<JJ>*<NN.*>}'
        self.chunker = nltk.RegexpParser(pattern)
        self.nlp = spacy.load('en_core_web_sm')

    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        question = example_dict['question']
        entities = self._extract_entities(question)
        if len(entities) == 0:
            raise RuntimeError(
                "SemanticChangeSteps failed to extract any entities from "
                + "the question "
                + str(question)
            )
        steps = example_dict['steps']
        if len(steps) == 0:
            raise RuntimeError(
                "SemanticChangeSteps requires at least 1 step. Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )

        result = copy.deepcopy(steps)
        changed_steps = []
        for i, step_to_dupe in enumerate(result):
            step_entities = self._extract_entities(step_to_dupe)
            if len(step_entities) == 0:
                continue
            # we want to sample an entity from the step and replace it with
            # another entity from the question. So we first sample from
            # step_entities, and if there are items in the question entities
            # that are not equal to it, then we use the diff to sample
            # the entity to replace it with. However, if the sampled entity
            # is the only entity in the question entities, we try to sample
            # another item from step_entities if there are more items.
            entity_to_replace = self.random.choice(step_entities)
            if (
                len(entities) == 1
                and entities[0] == entity_to_replace
                and len(step_entities) > 0
            ):
                entity_to_replace = self.random.choice(
                    [se for se in step_entities if se != entity_to_replace]
                )
            entities_to_choose_from = [
                ent for ent in entities if ent != entity_to_replace
            ]
            if len(entities_to_choose_from) == 0:
                continue
            chosen_entity = self.random.choice(entities_to_choose_from)
            changed_step = step_to_dupe.lower().replace(
                entity_to_replace, chosen_entity, 1
            )
            changed_steps.append((i, changed_step))
        if not changed_steps:
            raise RuntimeError(
                "SemanticChangeStep failed for the reasoning chain "
                + " ".join(str(step) for step in steps)
                + ". It is likely that this perturbation cannot be applied"
                + " to your dataset."
            )

        to_dupe = self.random.choice(range(len(changed_steps)))
        result[changed_steps[to_dupe][0]] = changed_steps[to_dupe][1]
        example_dict['steps'] = result
        return example_dict

    def _extract_entities(self, text: str) -> List[str]:
        entities = set()
        try:
            # first use spacy
            named_entities = self.nlp(text)
            entities.update(set([X.text for X in named_entities.ents]))
            # next use nltk pos tagger
            text_pos = nltk.pos_tag(nltk.word_tokenize(text))
            chunks = self.chunker.parse(text_pos)
            for subtree in chunks:
                if type(subtree) == nltk.tree.Tree:
                    entities.add(' '.join(t[0] for t in subtree))
        except TypeError:
            pass
        except ValueError:
            pass
        return [entity.lower() for entity in entities]


PERTURBATIONS_LIST = [
    ShuffleSteps,
    DuplicateOneStep,
    RemoveOneStep,
    SwapOneStep,
    ExtrinsicHallucinatedStep,
    ParaphraseSteps,
    GrammaticalErrorStep,
    NegateStep,
    SemanticChangeStep,
]

PERTURBATIONS = {x.__name__: x for x in PERTURBATIONS_LIST}


class MWPStep:
    """
    Class to represent Math Word Problem Steps.

    Receives the input step string and extracts the numbers and mathematical operations
    in the step
    """

    def __init__(self, step: str):
        self.step = step
        self.operations = self.extract_operations()
        self.numbers = self.extract_numbers()

    def extract_operations(self) -> List[str]:
        """
        Finds all instances of the math operations: -, +, *, ^, / in the step.
        """
        if not self.step:
            return []
        try:
            operations = re.findall(r'[-+*^/]', self.step)
        except TypeError as e:
            print(f"TYPE: {type(self.step)}")
            print(f"STEP: {self.step}")
            raise e
        return operations

    def extract_numbers(self) -> List[str]:
        """
        Finds all instances of numbers in the step.
        """
        if not self.step:
            return []
        numbers = re.findall(r'[0-9]+', self.step)
        return numbers

    def get_operations(self) -> List[str]:
        return self.operations

    def get_numbers(self) -> List[str]:
        return self.numbers

    def __str__(self) -> str:
        return self.step

    def __repr__(self) -> str:
        return f'MWPStep("{self.step}")'


class MWPPerturbation(StepPerturbation):
    def steps_empty_check(self, example_dict=Dict):
        mwp_steps = example_dict["steps"]
        if len(mwp_steps) == 0:
            raise RuntimeError(
                "MWPPerturbation requires at least 1 step. Make sure `get_data_for_fold()`"
                + "appropriately filters for these turns when used with this perturbation"
            )

    def perform_replacements(
        self, mwp_step: MWPStep, old_elements: List[str], new_elements: List[str]
    ) -> MWPStep:
        """
        Replaces the old_elements in self.step with the new_elements in the order in
        which they come.

        old_elements : List[str]
            list of the numbers or operations currently in the step in the order in which they appear
        new_elements : List[str]
            list of the numbers or operations to substitute in the step in the order in which they appear
        """

        remaining_old_step = mwp_step.step
        new_step_string = ""
        for old_element, new_element in zip(old_elements, new_elements):
            old_element_idx = remaining_old_step.find(old_element)
            new_step_string += remaining_old_step[:old_element_idx]
            new_step_string += new_element
            remaining_old_step = remaining_old_step[
                old_element_idx + len(old_element) :
            ]
        new_step_string += remaining_old_step
        return MWPStep(new_step_string)


class ShuffleNumbers(MWPPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        self.steps_empty_check(example_dict)
        result = []
        shuffled = False
        for mwp_step in example_dict["steps"]:
            numbers = mwp_step.get_numbers()
            # at least 2 distinct numbers
            if len(numbers) < 2 or all(
                [numbers[i] == numbers[i + 1] for i in range(len(numbers) - 1)]
            ):
                result.append(mwp_step.step)
                continue
            shuffled_numbers = copy.deepcopy(numbers)
            # make sure steps are different from original
            while " ".join(shuffled_numbers) == " ".join(numbers):
                shuffled_numbers = self.random.sample(numbers, len(numbers))
            perturbed_mwp_step = self.perform_replacements(
                mwp_step, numbers, shuffled_numbers
            )
            result.append(perturbed_mwp_step)
            shuffled = True
        if not shuffled:
            raise RuntimeError(
                "ShuffleNumbers requires at least 2 numbers in one of the steps.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        example_dict['steps'] = result
        return example_dict


class ShuffleOperations(MWPPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        self.steps_empty_check(example_dict)
        result = []
        shuffled = False
        for mwp_step in example_dict["steps"]:
            operations = mwp_step.get_operations()
            # at least 2 distinct operations
            if len(operations) < 2 or all(
                [operations[i] == operations[i + 1] for i in range(len(operations) - 1)]
            ):
                result.append(mwp_step.step)
                continue
            shuffled_operations = copy.deepcopy(operations)
            # make sure steps are different from original
            while shuffled_operations == operations:
                shuffled_operations = self.random.sample(operations, len(operations))
            perturbed_mwp_step = self.perform_replacements(
                mwp_step, operations, shuffled_operations
            )
            shuffled = True
            result.append(perturbed_mwp_step)
        if not shuffled:
            raise RuntimeError(
                "ShuffleOperations requires at least 2 operations in one of the steps.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        example_dict['steps'] = result
        return example_dict


class RandomNumber(MWPPerturbation):
    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        self.steps_empty_check(example_dict)
        result = []
        changed = False
        for mwp_step in example_dict["steps"]:
            numbers = mwp_step.get_numbers()
            if len(numbers) == 0:
                result.append(mwp_step.step)
                continue
            new_numbers = copy.deepcopy(numbers)
            while "".join(new_numbers) == "".join(numbers):
                rand_num = str(self.random.randint(0, 100))
                rand_num_idx = self.random.randint(0, len(numbers) - 1)
                new_numbers[rand_num_idx] = rand_num
            perturbed_mwp_step = self.perform_replacements(
                mwp_step, numbers, new_numbers
            )
            result.append(perturbed_mwp_step)
            changed = True
        if not changed:
            raise RuntimeError(
                "RandomNumber requires at least 1 number in the chain.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        example_dict['steps'] = result
        return example_dict


class RandomOperation(MWPPerturbation):
    def __init__(self, opt: Opt, cache: Optional[Dict[str, List[str]]] = None):
        super().__init__(opt, cache)
        self.random = random.Random(32)
        self.operations = ['+', '-', '/', '*', '^']

    def perturb(self, example_dict: Dict) -> Dict:
        example_dict = copy.deepcopy(example_dict)
        self.steps_empty_check(example_dict)
        result = []
        changed = False
        for mwp_step in example_dict["steps"]:
            operations = mwp_step.get_operations()
            if len(operations) == 0:
                result.append(mwp_step.step)
                continue
            new_operations = copy.deepcopy(operations)
            while "".join(new_operations) == "".join(operations):
                rand_operation = self.random.choice(self.operations)
                rand_operation_idx = self.random.randint(0, len(operations) - 1)
                new_operations[rand_operation_idx] = rand_operation
            perturbed_mwp_step = self.perform_replacements(
                mwp_step, operations, new_operations
            )
            result.append(perturbed_mwp_step)
            changed = True
        if not changed:
            raise RuntimeError(
                "RandomOperation requires at least 1 operation in the chain.  Make sure `get_data_for_fold`()`"
                + "appropriately filters for these turns when used with this perturbation."
            )
        example_dict['steps'] = result
        return example_dict


MATH_PERTURBATIONS_LIST = [
    ShuffleNumbers,
    ShuffleOperations,
    RandomNumber,
    RandomOperation,
]

MATH_PERTURBATIONS = {x.__name__: x for x in MATH_PERTURBATIONS_LIST}


class StepByStepReason(AbstractReason):
    """
    Class to represent step-by-step style reasoning.

    There are a few simple perturbations that we include by default. For tasks which
    have more complicated perturbations (ex. some way of modifying step-by-step
    reasoning that is also context dependent) it is highly recommended to EXTEND this
    class to a new base class that is specific for that dataset.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("StepByStepReason args")
        group.add_argument(
            "--step-by-step-style",
            choices=["scratchpad", "thought_prefix", "none"],
            default="thought_prefix",
            type=str,
            help="How do we display the step-by-step reasoning? Do we do it like scratchpad (with `scratch_start`"
            + "and `scratch_end` tokens) or with a prefix for the reasoning turns?",
        )
        group.add_argument("--thought-token", default="THINK: ", type=str)
        group.add_argument(
            "--scratch-start-token", default="__scratch_start__", type=str
        )
        group.add_argument("--scratch-end-token", default="__scratch_end__", type=str)
        group.add_argument(
            "--step-perturbations",
            default=[],
            nargs="*",
            help="Are we going to try to mess with any of the steps in this class?",
            type=str,
            choices=PERTURBATIONS.keys(),
        )
        return parser

    def __init__(self, opt):
        super().__init__(opt)
        self.cache_file_path = os.path.join(
            opt['datapath'], 'reasoning', 'step_by_step', 'perturbation_cache.json'
        )
        self.perturbation_cache = (
            json.load(open(self.cache_file_path))
            if os.path.exists(self.cache_file_path)
            else {}
        )
        self.perturbations = [
            PERTURBATIONS[p](self.opt, self.perturbation_cache.get(p, None))
            for p in self.opt.get("step_perturbations", [])
        ]

    def get_maybe_perturbed_steps(self, example_dict: Dict) -> List[str]:
        for perturbation in self.perturbations:
            example_dict = perturbation.perturb(example_dict)
        return example_dict['steps']

    def batch_get_maybe_perturbed_steps(
        self, batch_of_examples: List[Dict]
    ) -> List[List[str]]:
        for perturbation in self.perturbations:
            batch_of_examples = perturbation.batch_perturb(batch_of_examples)
        return [ex['steps'] for ex in batch_of_examples]

    def format_steps(self, steps_to_format):
        if self.opt.get("step_by_step_style") == "scratchpad":
            steps = (
                [self.opt["scratch_start_token"]]
                + steps_to_format
                + [self.opt["scratch_end_token"]]
            )
        elif self.opt.get("step_by_step_style") == "none":
            steps = [str(x) for x in steps_to_format]
        else:
            steps = [self.opt["thought_token"] + str(x) for x in steps_to_format]
        return steps

    def get_full_reason_text(self, example_dict) -> t_REASON_PROCESSED_RESULT:
        original_steps_raw = example_dict["steps"]
        processed_steps = self.format_steps(
            self.get_maybe_perturbed_steps(example_dict)
        )
        original_steps = self.format_steps(original_steps_raw)
        return (
            "",
            self.inner_separator.join(processed_steps),
            {
                "steps": processed_steps,
                "perturbations": self.opt.get("step_perturbations", []),
                "original_steps": original_steps,
            },
        )

    def batch_get_full_reason_text(
        self, examples: List[Dict]
    ) -> List[t_REASON_PROCESSED_RESULT]:
        original_steps = [self.format_steps(ex["steps"]) for ex in examples]
        processed_steps = self.batch_get_maybe_perturbed_steps(examples)
        processed_steps = [self.format_steps(s) for s in processed_steps]
        return [
            (
                "",
                self.inner_separator.join(ps),
                {
                    "steps": ps,
                    "perturbations": self.opt.get("step_perturbations", []),
                    "original_steps": os,
                },
            )
            for os, ps in zip(original_steps, processed_steps)
        ]

    def save_cache(self) -> None:
        os.makedirs(os.path.dirname(self.cache_file_path), exist_ok=True)
        with open(self.cache_file_path, 'w') as dest:
            for name, p in zip(
                self.opt.get("step_perturbations", []), self.perturbations
            ):
                self.perturbation_cache[name] = p.cache
            json.dump(
                self.perturbation_cache,
                dest,
                default=(lambda x: list(x) if isinstance(x, set) else x),
            )


class MWPStepsReason(StepByStepReason):
    """
    Class to represent step-by-step style reasoning for math datasets.

    Replaces example_dict["steps"] with MWPSteps to support functionality pertaining to
    numbers and mathematical operations contained within the steps
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("MWPStepsReason args")
        group.add_argument(
            "--math-step-perturbations",
            default=[],
            nargs="*",
            help="Are we going to try to mess with any of the math steps in this class?",
            type=str,
            choices=MATH_PERTURBATIONS.keys(),
        )
        return parser

    def __init__(self, opt):
        super().__init__(opt)
        for perturbation in self.opt.get("math_step_perturbations", []):
            self.perturbations.append(MATH_PERTURBATIONS[perturbation](self.opt))

    def get_full_reason_text(self, example_dict) -> t_REASON_PROCESSED_RESULT:
        out_dict = copy.deepcopy(example_dict)
        if len(self.perturbations) > 0:
            out_dict["steps"] = []
            for step in example_dict["steps"]:
                if not step:
                    continue
                if type(step) is str:
                    out_dict["steps"].append(MWPStep(step))
                elif type(step) is MWPStep:
                    out_dict["steps"].append(step)
                else:
                    raise ValueError(f"Unknown step type: {type(step)} for step {step}")
        return super().get_full_reason_text(out_dict)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

from parlai.core.message import Message
from parlai.core.mutators import MessageMutator, register_mutator
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.data import DatatypeHelper
import hashlib
import parlai.utils.logging as logging

import copy


"""
This mutator introduces noise to the training data by either fliping the labels or random set the labels,
work with binary classification only.
"""


@register_mutator("flip_classification_label")
class FlipClassificationLabelMutator(MessageMutator):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('FlipLabelMutator args')
        agent.add_argument(
            '--noise-level',
            type=float,
            help='The probability of label flipping',
        )
        agent.add_argument(
            '--classification-label-to-mutate',
            default='__ok__,__notok__',
            type=str,
            help='The binary labels in the data for flipping / random assignment. seprated by ,',
        )
        return parser

    def __init__(self, opt: Opt):
        super().__init__(opt)
        if opt['noise_level'] > 1 or opt['noise_level'] < 0:
            raise Exception('Noise_level should be within 0 - 1.')
        if opt['noise_level'] == 0:
            logging.info('No noise, teacher should be the same as before.')
        if opt['noise_level'] == 1:
            logging.info('Labels are exact the opposite.')
        self.labels = opt['classification_label_to_flip'].split(',')
        self.noise_level = opt['noise_level']
        logging.info(f'Flipping with noise level {self.noise_level}')
        self.is_training = DatatypeHelper.is_training(opt['datatype'])
        self.flip_cnt = 0
        return

    def message_mutation(self, message: Message) -> Message:
        message = copy.deepcopy(message)
        if 'labels' in message:
            labels = message['labels']
            text = message['text']
            new_labels = []
            flip = hashlib.md5(text.encode('utf8')).digest()[0] < self.noise_level * 256
            if len(labels) != 1:
                raise ValueError(
                    f'{type(self).__name__} can only be used with one label!'
                )
            else:
                # deterministic flipping for each example
                if flip:
                    self.flip_cnt += 1
                    if labels[0] == self.labels[0]:
                        new_labels = [self.labels[1]]
                    elif labels[0] == self.labels[1]:
                        new_labels = [self.labels[0]]
                    else:
                        raise ValueError(
                            f'labels must be binary and the same as in {self.labels}'
                        )
                    assert len(new_labels) == 1
            if flip:
                message.force_set('labels', new_labels)
        return message

    def __del__(self):
        logging.info(f'Flipped {self.flip_cnt}')


@register_mutator("flip_classification_label_train_only")
class FlipClassificationLabelTrainOnlyMutator(FlipClassificationLabelMutator):
    def __init__(self, opt: Opt):
        super().__init__(opt)
        if not self.is_training:
            logging.info('No flipping in eval / test mode.')
        return

    def message_mutation(self, message: Message) -> Message:
        if not self.is_training:
            return message
        else:
            return super().message_mutation(message)


@register_mutator("flip_classification_label_valid_only")
class FlipClassificationLabelValidOnlyMutator(FlipClassificationLabelMutator):
    def __init__(self, opt: Opt):
        super().__init__(opt)
        if self.is_training:
            logging.info('No flipping in train mode.')
        return

    def message_mutation(self, message: Message) -> Message:
        if self.is_training:
            return message
        else:
            return super().message_mutation(message)


@register_mutator("random_classification_label")
class RandomClassificationLabelMutator(FlipClassificationLabelMutator):
    def message_mutation(self, message: Message) -> Message:
        message = copy.deepcopy(message)
        if 'labels' in message:
            labels = message['labels']
            text = message['text']
            new_labels = []
            flip = hashlib.md5(text.encode('utf8')).digest()[0] < self.noise_level * 256
            if len(labels) != 1:
                raise ValueError(
                    f'{type(self).__name__} can only be used with one label!'
                )
            else:
                # deterministic flipping for each example
                if flip:
                    self.flip_cnt += 1
                    if hashlib.md5(text.encode('utf8')).digest()[1] / 256 > 0.5:
                        new_labels = [self.labels[0]]
                    else:
                        new_labels = [self.labels[1]]

                    assert len(new_labels) == 1
            if flip:
                message.force_set('labels', new_labels)
        return message



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

# Pin against a specific commit since folks occasionally post fixes
MULTIWOZ_URL_BASE = "https://raw.githubusercontent.com/budzianowski/multiwoz/01e689362833ce33427a771a21cefe253e8f5886/"

MULTIWOZ_22_URL_BASE = MULTIWOZ_URL_BASE + "data/MultiWOZ_2.2/"

RESOURCES = [
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'dialog_acts.json',
        'dialog_acts.json',
        '328f392165e7826db9f827731b14b5cc04e79e9e3c6332bfb192a1ea17f8e9b6',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'schema.json',
        'schema.json',
        'ae9e2390f38fb967af64623c2f4f7e0c636fb377ad523b582a03161d3ddbdf68',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'dev/dialogues_001.json',
        'dev/dialogues_001.json',
        'e7ddb563e4da5766ea820cc826dead77e7ca219c19b761e218d62d9c999a252e',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'dev/dialogues_002.json',
        'dev/dialogues_002.json',
        'ede6a2c17fd6c5846214b8cabc1ef8f7cc8be01cfbacaa162bcafec9e87724e9',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'test/dialogues_001.json',
        'test/dialogues_001.json',
        'd6f43876cf130fdb2dfa8f96bc056b0995354137f02122e004925d01264ed386',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'test/dialogues_002.json',
        'test/dialogues_002.json',
        '89af95d8f596a448e733d59b31be78f1dd1632eddd99d5cb298a3fcb1ac9d185',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_001.json',
        'train/dialogues_001.json',
        '895a8109bf01fa5ecf15ccdbd2dfe1628bd923f6b61dcd2e26b10ee5076a1596',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_002.json',
        'train/dialogues_002.json',
        '2f3ea771d4e01cb2780357738cff7f7496b87d34c221cc240df74501312438d3',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_003.json',
        'train/dialogues_003.json',
        'da24961d28486be2d8462ee4d86a809db819d588ba90ae1a783383d95eb85daa',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_004.json',
        'train/dialogues_004.json',
        '30c1172db1071c853b16215d1946de908d68d2b6ff8da7801de307000a179106',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_005.json',
        'train/dialogues_005.json',
        'eaf58716df5de99524b3e0e7edf477b74749512788a6a51f53f2bdd76768d39a',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_006.json',
        'train/dialogues_006.json',
        '8e75fd543b1964bc5e7118085d977f479c98fcdf6d606b971f67a45fb1745c83',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_007.json',
        'train/dialogues_007.json',
        '02323f8298439d713c6d7d226f4bd7ec246ec993ee11911b54f98cb8a598f206',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_008.json',
        'train/dialogues_008.json',
        '1129fbed480352ae304f0ae5b4915c194e9619c43f1577ccb0d450e10d24ea98',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_009.json',
        'train/dialogues_009.json',
        '87d9e43b1ba51a4a4688703da79d3a09b14d8013570545da24c597daa18e2f45',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_010.json',
        'train/dialogues_010.json',
        'e7ad0d5da2909b08197295e45fe4695b9dc2f67d458374b2aab8db5094e97b26',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_011.json',
        'train/dialogues_011.json',
        '82e2d2900a037b866a01d05974734dd419e89b329fe29ef93b35eea96d27feb8',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_012.json',
        'train/dialogues_012.json',
        'b6bf292325db67682dd7b6fafbf1051cc2262e92f7c37cab213e975888594bb2',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_013.json',
        'train/dialogues_013.json',
        'c33fe4b3952c016e1e1645f472a7097f93dfb476a19940fd9386865ef9adf685',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_014.json',
        'train/dialogues_014.json',
        'ce33dbbf93a40d0dcc671a9d6a2ed1f987914f5b5f05f6df493a5b342efda954',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_015.json',
        'train/dialogues_015.json',
        'd895c0439bc2ad89ef1689896e3be630eadebc33dae77b42a426fd16a271718e',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_016.json',
        'train/dialogues_016.json',
        '3e6bc0bca4262022ccbce0d5ce3150e536e7d21aeb9fbdef9e83cced4dfd124b',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_22_URL_BASE + 'train/dialogues_017.json',
        'train/dialogues_017.json',
        'b6ab2cd9b6e8983364526845b7cbec1749338209bf7ac2313c25e1e2226ebab5',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/attraction_db.json',
        'db/attraction_db.json',
        '2aacc620af4025f1eada5ec83057a866f8e8b72b529f71d2f8bf93bcdd8f8751',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/bus_db.json',
        'db/bus_db.json',
        '4818e735bae20690f6d0d06bb2ae8eec1981c0b680258a970dc01c9073f3fec9',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/hospital_db.json',
        'db/hospital_db.json',
        'f28738bda15e750be912d653c5e68b06af41dba68d9cfa3febfdcfe972e14366',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/hotel_db.json',
        'db/hotel_db.json',
        '972bbd65beada7c64f0b87322c694fd9173b46cf8e61ca3bbe951717ac1d1662',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/police_db.json',
        'db/police_db.json',
        'd9c2b200fa2dd61b04ce2fe520b0b79dfa68d8b895806cfec6e8a8d3cffa9193',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/restaurant_db.json',
        'db/restaurant_db.json',
        '7864b4e36027af0907d6afe9ed962fecd6c6966cd626b8ae9341708a04ea201a',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/taxi_db.json',
        'db/taxi_db.json',
        '08b8fb2436abec6d1fe9087054f943d2b31e1c4adc74d8202e4e2149a5630be3',
        zipped=False,
    ),
    DownloadableFile(
        MULTIWOZ_URL_BASE + 'db/train_db.json',
        'db/train_db.json',
        '4818e735bae20690f6d0d06bb2ae8eec1981c0b680258a970dc01c9073f3fec9',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'multiwoz_v22')
    version = '1.0'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        build_data.make_dir(dpath + "/dev")
        build_data.make_dir(dpath + "/test")
        build_data.make_dir(dpath + "/train")

        build_data.make_dir(dpath + "/db")

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Agents (used for dumping data) and Teachers (for training models) related to the TOD
conversation setup.

As a convention, agents and teachers that are inheritable are prefixed with "Tod"
whereas those that can be used as-is are not. Similarly, classes and functions that do
not need to be exposed outside of this file are prefixed with a single underscore ('_')
"""

from parlai.core.agents import Agent
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.utils.data import DatatypeHelper
from parlai.utils.distributed import is_distributed, get_rank, num_workers

import parlai.core.tod.tod_core as tod
from parlai.core.tod.tod_core import SerializationHelpers
from parlai.core.tod.teacher_metrics import SlotMetrics, NlgMetrics

from typing import Optional, List
import json
import pickle
import difflib
import random
from math import ceil


######### Agents that dump information from a dataset; base classes
class TodStructuredDataParser(Agent):
    """
    Base class that specifies intermediate representations for Tod conversations.

    Inherit from this class and implement `setup_episodes()` to implement the intermediate representation for a specific dataset. Use multiple inheritence with classes that implement an `act()` below to use.

    For example, if we have a `MyDataset_DataParser(TodStructuredDataParser)` and wanted to make a teacher to train a model to generate User Utterances based on a goal prompt, we would do so by defining `class MyDatasetUserSimulatorTeacher(MyDataset_DataParser, TodUserSimulatorTeacher)`.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        if hasattr(super(), "add_cmdline_args"):
            parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("TOD StructuredData agent")
        group.add_argument(
            "--episodes-randomization-seed",
            type=int,
            default=-1,
            help="Randomize episodes in a predictable way (eg, for few shot). Set to -1 for no randomization. ",
        )
        parser.add_argument(
            "--n-shot",
            default=-1,
            type=int,
            help="Number of dialogues to keep for each of train/valid/test. -1 means all. Dialogues of lower numbers are strict subsets of larger numbers. Do not use in conjunction with `--percent-shot`. Use `--episodes-randomization-seed` to change seed. NOTE: Beware of using this flag when multitasking as this will apply to *all* datasets unless the ':' syntax for specifying per-dataset flags is used.",
        )
        parser.add_argument(
            "--percent-shot",
            default=-1,
            type=float,
            help="Percentage of dialogues to keep for each of train/valid/test. -1 means all. Dialogues of lower numbers are strict subsets of larger numbers. Do not use in conjunction with `--n-shot`. Use `--episodes-randomization-seed` to change seed. NOTE: Beware of using this flag when multitasking as this will apply to *all* datasets unless the ':' syntax for specifying per-dataset flags is used.",
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = self.get_id_task_prefix() + "_" + self._get_agent_type_suffix()
        if shared is None:
            self.episodes = self.generate_episodes()
        else:
            self.episodes = shared["episodes"]

    def share(self):
        share = super().share()
        share["episodes"] = self.episodes
        return share

    def setup_episodes(self, fold: str) -> List[tod.TodStructuredEpisode]:
        """
        Fold here is a data fold.
        """
        raise NotImplementedError(
            "Must have method for generating an episode. Must be set in downstream Parser for a given task"
        )

    def generate_episodes(self) -> List[tod.TodStructuredEpisode]:
        if self.opt.get("n_shot", -1) >= 0 and self.opt.get("percent_shot", -1) >= 0:
            # Validate before spending a while to load eeverything
            raise RuntimeError("Both `--n-shot` and `--percent-shot` in use!")
        episodes = list(self.setup_episodes(self.fold))
        if self.opt.get("episodes_randomization_seed", -1) != -1:
            random.Random(self.opt["episodes_randomization_seed"]).shuffle(episodes)
        if self.opt.get("n_shot", -1) != -1:
            episodes = episodes[: self.opt["n_shot"]]
        elif self.opt.get("percent_shot", -1) >= 0:
            episodes = episodes[: int(len(episodes) * self.opt["percent_shot"])]
        return episodes

    def get_id_task_prefix(self) -> str:
        """
        Convenience for setting IDs.
        """
        raise NotImplementedError(
            "Must set ID prefix in downstream task agent. Must be set in downsream Parser for a given task"
        )

    def _get_agent_type_suffix(self) -> str:
        """
        Convenience for setting IDs.
        """
        raise NotImplementedError(
            "Must set in downstream agent within `tod_agents`. If you see this error, something is wrong with TOD Infrastructure"
        )


######### Agents that dump information from a dataset as gold (explicitly should *not* be used with teachers)
class _TodDataDumpAgent(TodStructuredDataParser):
    """
    For agents which dump data from some dataset, without training/other modifications.

    Since we have to deal with batching inside of agents (as per ParlAI convention for
    non-generative agents), this does so while also implementing an "epoch done" to
    denote elements in a batch that are past the end of the epoch.
    """

    def __init__(self, opt: Opt, shared=None):
        if not hasattr(self, "fold"):
            self.fold = DatatypeHelper.fold(opt["datatype"])
        super().__init__(opt, shared)
        self.epochDone = False
        self.batchsize = opt.get("batchsize", 1)
        self.max_episodes = len(self.episodes)
        if opt.get("num_episodes", 0) > 0:
            self.max_episodes = min(self.max_episodes, opt.get("num_episodes"))
        self.episode_idx = opt.get("batchindex", 0)
        self._setup_next_episode()
        self.round_idx = 0  # for some downstream utt + sysUttAndApiCallAgents.
        if is_distributed():  # cause gotta manually handle
            rank = get_rank()
            chunk_size = ceil(self.max_episodes / num_workers())
            self.episode_idx += rank * chunk_size
            self.max_episodes = min(self.max_episodes, (rank + 1) * chunk_size)

    def _setup_next_episode(self):
        self.epochDone = self.episode_idx >= self.max_episodes
        self.episode = None
        if not self.epochDone:
            self.episode = self.episodes[self.episode_idx]
        self.round_idx = (
            0  # so downstream agents know which round they are in. Update in `act()`
        )

    def epoch_done(self) -> bool:
        return self.epochDone

    def episode_done(self) -> bool:
        raise RuntimeError("Must be defined in downstream agent")

    def num_episodes(self) -> int:
        return len(self.episodes)

    def reset(self):
        self.episode_idx += self.batchsize
        self._setup_next_episode()


class TodGoalAgent(_TodDataDumpAgent):
    """
    Use as a mixin with a dataset parser class that includes `generate_episodes()` of
    TodStructuredDataParser.

    Dumps out all goal calls from an episode.
    """

    def act(self):
        return {
            "text": f"{tod.STANDARD_GOAL}{self.episode.goal_calls_utt}",
            "id": self.id,
            "domain": self.episode.domain,
            "episode_done": False,
        }

    def _get_agent_type_suffix(self):
        return "Goal"

    def episode_done(self) -> bool:
        # done if end of batch; should never end conversation otherwise
        return self.epoch_done()


class TodApiSchemaAgent(_TodDataDumpAgent):
    """
    Use as a mixin with a dataset parser class that includes `generate_episodes()` of
    TodStructuredDataParser.

    Dumps out api schemas associated with an episode, based on what is manually set in
    the dataset parser.
    """

    def act(self):
        return {
            "text": f"{tod.STANDARD_API_SCHEMAS}{self.episode.api_schemas_utt}",
            "id": self.id,
            "domain": self.episode.domain,
            "episode_done": False,
        }

    def _get_agent_type_suffix(self):
        return "ApiSchema"

    def episode_done(self) -> bool:
        # done if end of batch; should never end conversation otherwise
        return self.epoch_done()


############# Single Goal + Api Schema Agent
class _EpisodeToSingleGoalProcessor(_TodDataDumpAgent):
    """
    Iterate through all of the goals of a dataset, one by one.

    Slightly different logic than the dump agent since how we count + setup examples for
    an episode are different

    Used as a mixin in the SingleGoal and SingleApiSchema agents below.

    This class exposes a `filter_goals()` function that can be overridden by downstream agents.
    """

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.epochDone = False
        if shared is None:
            self.episodes = self._setup_single_goal_episodes()
        else:
            # Handled fine in _TodDataDumpAgent
            pass

        self.max_episodes = len(self.episodes)
        if opt.get("num_episodes", 0) > 0:
            self.max_episodes = min(self.max_episodes, opt.get("num_episodes"))
        if is_distributed():  # cause gotta manually handle
            rank = get_rank()
            chunk_size = ceil(self.max_episodes / num_workers())
            self.max_episodes = min(self.max_episodes, (rank + 1) * chunk_size)

        self._setup_next_episode()

    def _setup_single_goal_episodes(self) -> List[tod.TodStructuredEpisode]:
        """
        This function assumes that `self.setup_episodes()` has already been called
        prior.

        Based on the `__init__` order of this class, it should be done in
        `TodStructuredDataParser` by this point.
        """
        raw_episodes = self.episodes
        result = []
        for raw in raw_episodes:
            for call in self.filter_goals(raw.goal_calls_machine):
                schema = {}
                for cand in raw.api_schemas_machine:
                    if (
                        cand[tod.STANDARD_API_NAME_SLOT]
                        == call[tod.STANDARD_API_NAME_SLOT]
                    ):
                        schema = cand

                result.append(
                    tod.TodStructuredEpisode(
                        domain=raw.domain,
                        api_schemas_machine=[schema],
                        goal_calls_machine=[call],
                        rounds=[],
                    )
                )
        return result

    def filter_goals(self, goals):
        """
        Some downstream agents may want to filter the goals.

        Override this if so.
        """
        return goals


class TodSingleGoalAgent(_EpisodeToSingleGoalProcessor, TodGoalAgent):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Takes goals of an episode and splits them into single versions. (That is, if an episode has 3 goal API calls, this makes it such that those 3 goal API calls become the grounding for 3 separate episodes.)

    NOTE: If an API schema agent is used, this *must* be used with `TodSingleApiSchemaAgent` since it will be nonsensicle otherwise. Additionally, this agent will not function properly with UserUtt + SystemUttAndApiCall agent, since episodes will not align.
    """

    def _get_agent_type_suffix(self):
        return "SingleGoal"


class TodSingleApiSchemaAgent(_EpisodeToSingleGoalProcessor, TodApiSchemaAgent):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Takes the schema provided for an episode and filters these to match the single Goal provided by TodSingelGoalAgent.

    NOTE: Must be used with TodSingleGoalAgent since nonsensicle otherwise. Additionally, this agent will not function properly with UserUtt + SystemUttAndApiCall agent, since episodes will not align.
    """

    def _get_agent_type_suffix(self):
        return "SingleApiSchema"


###### Agents used for calculating TOD World Metrics based on a dataset. See `tod_world_script` or `parlai/projects/tod_simulator/` for examples.
class TodUserUttAgent(_TodDataDumpAgent):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Agent provided as a convenience to run TOD World script code on a dataset without having to write too much code to do so. (Ex. for a quick way to dump data to a `.jsonl` file for generating data for ACUTE or to generate a report file of metrics from TodWorld script.)

    This represents the "User" agent.

    This class should only ever be used with the model-model chat world which will stop
    upon seeing the '[DONE]' utterance; may go out of bounds otherwise.
    """

    def act(self):
        result = {
            "text": f"{tod.STANDARD_USER_UTTERANCE}{self.episode.rounds[self.round_idx].user_utt}",
            "id": self.id,
            "domain": self.episode.domain,
            "episode_done": False,
        }
        self.round_idx += 1
        return result

    def reset(self):
        super().reset()  # setup next episode
        self.round_idx = 0

    def _get_agent_type_suffix(self):
        return "User"

    def episode_done(self) -> bool:
        return self.epoch_done() or self.round_idx >= len(self.episode.rounds)


class TodApiCallAndSysUttAgent(_TodDataDumpAgent):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Agent provided as a convenience to run TOD World script code on a dataset without having to write too much code to do so. (Ex. for a quick way to dump data to a `.jsonl` file for generating data for ACUTE or to generate a report file of metrics from TodWorld script.)

    This class represents the System and will generate both API Calls and System Utterances.

    This class should only ever be used with the model-model chat world which will stop
    upon seeing the '[DONE]' utterance; may go out of bounds otherwise.
    """

    def __init__(self, opt: Opt, shared=None):
        # This class will have `act()` called on it twice per round — once for API call and once for NLG — so need to make sure we don't increment episode number (reset) prematurely; use the `already_reset` flag for this.
        self.already_reset = False
        self.api_call_turn = True
        super().__init__(opt, shared)

    def act(self):
        self.already_reset = False
        if tod.STANDARD_API_SCHEMAS in self.observation.get("text", ""):
            return {
                "text": tod.STANDARD_API_SCHEMAS,  # Default convention for the first turn
                "id": self.id,
                "domain": self.episode.domain,
                "episode_done": False,
            }

        if self.api_call_turn:  # comes first, don't iterate round #
            result = {
                "text": f"{tod.STANDARD_CALL}{self.episode.rounds[self.round_idx].api_call_utt}",
                "id": self.id,
                "domain": self.episode.domain,
                "episode_done": False,
            }
            self.api_call_turn = False
        else:
            result = {
                "text": f"{tod.STANDARD_SYSTEM_UTTERANCE}{self.episode.rounds[self.round_idx].sys_utt}",
                "id": self.id,
                "domain": self.episode.domain,
                "episode_done": False,
            }
            self.round_idx += 1
            self.api_call_turn = True

        return result

    def reset(self):
        if not self.already_reset:
            super().reset()  # setup next episode
            self.api_call_turn = True
            self.already_reset = True

    def _get_agent_type_suffix(self):
        return "System"

    def episode_done(self) -> bool:
        return self.epoch_done() or self.round_idx >= len(self.episode.rounds)


class TodApiResponseAgent(_TodDataDumpAgent):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Agent provided as a convenience to run TOD World script code on a dataset without having to write too much code to do so. (Ex. for a quick way to dump data to a `.jsonl` file for generating data for ACUTE or to generate a report file of metrics from TodWorld script.)

    This class represents the Api Response mechanism.

    This class should only ever be used with the model-model chat world which will stop
    upon seeing the '[DONE]' utterance; may go out of bounds otherwise.
    """

    def act(self):
        if tod.STANDARD_API_SCHEMAS in self.observation.get("text", ""):
            return {
                "text": tod.STANDARD_API_SCHEMAS,  # Default convention
                "id": self.id,
                "domain": self.episode.domain,
                "episode_done": False,
            }

        result = {
            "text": f"{tod.STANDARD_RESP}{self.episode.rounds[self.round_idx].api_resp_utt}",
            "id": self.id,
            "domain": self.episode.domain,
            "episode_done": False,
        }
        self.round_idx += 1
        return result

    def reset(self):
        super().reset()  # setup next episode
        self.round_idx = 0

    def _get_agent_type_suffix(self):
        return "ApiResponse"

    def episode_done(self) -> bool:
        return self.epoch_done() or self.round_idx >= len(self.episode.rounds)


###### Standalone API agent
class StandaloneApiAgent(Agent):
    """
    Trainable agent that saves API calls and responses.

    Use `TodStandaloneApiTeacher` to train this class. For example for a MultiWoz V2.2
    standalone API, use ``` parlai train -t multiwoz_v22:StandaloneApiTeacher -m
    parlai.core.tod.tod_agents:StandaloneApiAgent -eps 4 -mf output ``` to generate the
    `.pickle` file to use.
    """

    EMPTY_RESP = {
        "text": tod.STANDARD_RESP,
        "id": "StandaloneApiAgent",
        "episode_done": False,
    }

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        group = parser.add_argument_group("TOD Standalone API args")
        group.add_argument(
            "--exact-api-call",
            type=bool,
            default=True,
            help="Validation-time flag. If true, will return '' if exact api call values not found. If false, will pick response from the same intent with similar api parameters (assuming intent is the same when available)",
        )

        group.add_argument(
            "--fail-hard",
            type=bool,
            default=False,
            help="Aids in deugging. Will throw exception if API call not found and '--exact-api-call' is set.",
        )

        group.add_argument(
            "--standalone-api-file",
            type=str,
            default=None,
            help="Path to file holding `.pickle` of standalone api for validation (will intelligently strip if suffix included). If not set, assumes the `model_file` argument will contain the `.pickle` file. ",
        )
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id = "StandaloneApiAgent"
        file_key = "model_file"
        if self.opt["standalone_api_file"] is not None:
            file_key = "standalone_api_file"
        self.path_base = self.opt[file_key].replace(".pickle", "")
        self.db_path = self.path_base + ".pickle"
        self.exact_api_call = self.opt["exact_api_call"]
        try:
            with (open(self.db_path, "rb")) as openfile:
                self.data = pickle.load(openfile)
                self.training = True
                print("Loaded Standalone API data successfully")
            if self.exact_api_call != self.data.get("exact_api_call", True):
                raise RuntimeError(
                    f"Standalone API .pickle file generated with `exact_api_call` of {self.data.get('exact_api_call', False)} but StandaloneApiAgent sets it to {self.exact_api_call}"
                )
        except Exception:
            print(f"No file at {self.db_path}; ASSUMING WE ARE TRAINING")
            self.data = {}
            self.data["exact_api_call"] = self.exact_api_call
            self.training = True

    def _maybe_filter_prefix(self, text, prefix):
        if prefix in text:
            return text[len(prefix) :].strip()
        return text.strip()

    def act(self):
        if not self.observation["text"].startswith(tod.STANDARD_CALL):
            return self.EMPTY_RESP
        call_text_raw = self.observation["text"]
        # decode then reencode the API call so that we get the API calls in a consistent order
        call_text = SerializationHelpers.api_dict_to_str(
            SerializationHelpers.str_to_api_dict(
                call_text_raw[len(tod.STANDARD_CALL) :]
            )
        )
        if "labels" in self.observation:
            return self._do_train(call_text)
        return self._do_fetch(call_text)

    def _do_train(self, call_text):
        assert self.training is True
        self.data[call_text] = self.observation["labels"][0]
        return self.EMPTY_RESP

    def _do_fetch(self, call_text):
        if self.exact_api_call:
            if self.opt.get("fail_hard", False):
                resp = self.data[call_text]
            else:
                resp = self.data.get(call_text, tod.STANDARD_RESP)
            return {"text": resp, "id": self.id, "episode_done": False}

        # Not exact case
        best_key = difflib.get_close_matches(call_text, self.data.keys(), 1)
        if len(best_key) == 0:
            return self.EMPTY_RESP
        return {
            "text": self.data.get(best_key[0], tod.STANDARD_RESP),
            "id": self.id,
            "episode_done": False,
        }

    def shutdown(self):
        if self.training:
            with (open(self.db_path, "wb")) as openfile:
                pickle.dump(self.data, openfile)
                print(f"Dumped output to {self.db_path}")
            with open(self.path_base + ".opt", "w") as f:
                json.dump(self.opt, f)


######### Empty agents
class EmptyApiSchemaAgent(Agent):
    def __init__(self, opt, shared=None):
        super().__init__(opt)
        self.id = "EmptyApiSchemaAgent"

    def act(self):
        msg = {
            "id": self.getID(),
            "text": tod.STANDARD_API_SCHEMAS,
            "episode_done": False,
        }
        return Message(msg)


class EmptyGoalAgent(Agent):
    def __init__(self, opt, shared=None):
        super().__init__(opt)
        self.id = "EmptyGoalAgent"

    def act(self):
        msg = {"id": self.getID(), "text": tod.STANDARD_GOAL, "episode_done": False}
        return Message(msg)


############# Teachers
class TodSystemTeacher(TodStructuredDataParser, DialogTeacher):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    TOD agent teacher which produces both API calls and NLG responses.

    First turn is API Schema grounding, which may be a an empty schema.
    Subsequent turns alternate between
        1. User utterance -> API Call
        2. API Response -> System Utterance
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            "--api-schemas",
            type="bool",
            default=False,
            help="Preempt first turn with intents + required/optional parameters as key/value for given domain",
        )
        parser.add_argument(
            "--api-jga-record",
            type=bool,
            default=True,
            help="Breaks out jga into individual api schemas",
        )
        parser.add_argument(
            "--domain-jga-record",
            type=bool,
            default=False,
            help="Breaks out jga into individual domains",
        )
        parser.add_argument(
            "--domain-nlg-record",
            type=bool,
            default=False,
            help="Breaks out nlg into individual domains",
        )
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self._num_examples_cache = sum([len(x.rounds) * 2 + 1 for x in self.episodes])
        self._num_episodes_cache = len(self.episodes)

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        resp = model_response.get("text")
        if not resp:
            return
        if teacher_action["type"] == tod.STANDARD_CALL:
            if resp.startswith(tod.STANDARD_CALL):
                resp = resp[len(tod.STANDARD_CALL) :]
            predicted = SerializationHelpers.str_to_api_dict(resp)
            domains = (
                [teacher_action["domain"]] if self.opt["domain_jga_record"] else []
            )

            metrics = SlotMetrics(
                teacher_slots=teacher_action["slots"],
                predicted_slots=predicted,
                prefixes=domains,
            ).report()
            for key, value in metrics.items():
                self.metrics.add(key, value)

            if self.opt["api_jga_record"] and len(teacher_action["slots"]) > 0:
                teacher = teacher_action["slots"]
                slots = list(teacher.keys())
                slots.remove(tod.STANDARD_API_NAME_SLOT)
                api_here = (
                    "api-"
                    + teacher[tod.STANDARD_API_NAME_SLOT]
                    + "--"
                    + "-".join(slots)
                )
                self.metrics.add(f"{api_here}/jga", AverageMetric(teacher == predicted))

        elif teacher_action["type"] == tod.STANDARD_SYSTEM_UTTERANCE:
            domains = (
                [teacher_action["domain"]] if self.opt["domain_nlg_record"] else []
            )
            metrics = NlgMetrics(guess=resp, labels=labels, prefixes=domains).report()
            for key, value in metrics.items():
                self.metrics.add(key, value)

    def setup_data(self, fold):
        for episode in self.generate_episodes():
            if self.opt.get("api_schemas"):
                schemas = episode.api_schemas_utt
            else:
                schemas = ""
            yield {
                "text": f"{tod.STANDARD_API_SCHEMAS}{schemas}",
                "label": f"{tod.STANDARD_API_SCHEMAS}",
                "domain": episode.domain,
                "type": tod.STANDARD_API_SCHEMAS,
                "slots": {},
            }, True
            for r in episode.rounds:
                yield {
                    "text": f"{tod.STANDARD_USER_UTTERANCE}{r.user_utt}",
                    "label": f"{tod.STANDARD_CALL}{r.api_call_utt}",
                    "domain": episode.domain,
                    "type": tod.STANDARD_CALL,
                    "slots": r.api_call_machine,
                }, False
                yield {
                    "text": f"{tod.STANDARD_RESP}{r.api_resp_utt}",
                    "label": f"{tod.STANDARD_SYSTEM_UTTERANCE}{r.sys_utt}",
                    "domain": episode.domain,
                    "slots": r.api_resp_machine,
                    "type": tod.STANDARD_SYSTEM_UTTERANCE,
                }, False

    def _get_agent_type_suffix(self):
        return "SystemTeacher"


class TodUserSimulatorTeacher(TodStructuredDataParser, DialogTeacher):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Teacher that has `Goal->User Utterance` for its first turn, then `System
    Utterance->User Utterance` for all subsequent turns.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        # Manually set number of examples + number of episodes
        self._num_examples_cache = sum([len(x.rounds) for x in self.episodes])
        self._num_episodes_cache = len(self.episodes)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            "--api-schemas",
            type="bool",
            default=False,
            help="Preempt first turn with intents + required/optional parameters as key/value for given domain. NOOP for this teacher, but including to make sweeps easier",
        )
        return parser

    def setup_data(self, fold):
        for episode in self.generate_episodes():
            if len(episode.rounds) < 1:
                continue
            yield {
                "text": f"{tod.STANDARD_GOAL}{episode.goal_calls_utt}",
                "label": f"{tod.STANDARD_USER_UTTERANCE}{episode.rounds[0].user_utt}",
                "domain": episode.domain,
                "type": tod.STANDARD_USER_UTTERANCE,
            }, True
            for i, r in enumerate(episode.rounds):
                if i == len(episode.rounds) - 1:
                    continue
                yield {
                    "text": f"{tod.STANDARD_SYSTEM_UTTERANCE}{r.sys_utt}",
                    "label": f"{tod.STANDARD_USER_UTTERANCE}{episode.rounds[i+1].user_utt}",
                    "domain": episode.domain,
                    "type": tod.STANDARD_USER_UTTERANCE,
                    "slots": {},  # slots in agent/user turns are meaningless
                }, False

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        resp = model_response.get("text")
        if not resp:
            return
        if teacher_action["type"] == tod.STANDARD_RESP:
            if resp.startswith(tod.STANDARD_RESP):
                resp = resp[len(tod.STANDARD_RESP) :]
            predicted = SerializationHelpers.str_to_api_dict(resp)

            metrics = SlotMetrics(teacher_action["slots"], predicted).report()
            for key, value in metrics.items():
                self.metrics.add(key, value)

        elif teacher_action["type"] == tod.STANDARD_USER_UTTERANCE:
            metrics = NlgMetrics(resp, labels).report()
            for key, value in metrics.items():
                self.metrics.add(key, value)

    def _get_agent_type_suffix(self):
        return "UserSimulatorTeacher"


class TodStandaloneApiTeacher(TodStructuredDataParser, DialogTeacher):
    """
    Use as a mixin with classes that also extend + implement TodStructuredDataParser.

    Use this to generate a database for `StandaloneApiAgent`.

    Set this as the teacher with `StandaloneApiAgent` as the agent. Ex for a MultiWoz
    V2.2 standalone API, use ``` parlai train -t multiwoz_v22:StandaloneApiTeacher -m
    parlai.core.tod.tod_agents:StandaloneApiAgent -eps 4 -mf output ```
    """

    def setup_data(self, fold):
        # As a default, just put everything in
        for fold_overwrite in ["train", "valid", "test"]:
            for episode in self.setup_episodes(fold_overwrite):
                first = True
                for r in episode.rounds:
                    if len(r.api_call_machine) > 0:
                        yield {
                            "text": f"{tod.STANDARD_CALL}{r.api_call_utt}",
                            "label": f"{tod.STANDARD_RESP}{r.api_resp_utt}",
                            "id": self.id,
                            "domain": episode.domain,
                        }, first
                        first = False

    def _get_agent_type_suffix(self):
        return "StandaloneApiTeacher"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

ROOT_URL = (
    "https://raw.githubusercontent.com/xiul-msr/e2e_dialog_challenge/master/data/"
)

RESOURCES = [
    # raw data files
    DownloadableFile(
        f"{ROOT_URL}/movie_all.tsv",
        "movie_all.tsv",
        "d2291fd898d8c2d92d7c92affa5601a0561a28f07f6147e9c196c5a573a222d6",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/restaurant_all.tsv",
        "restaurant_all.tsv",
        "0e297b2ac2e29f9771fed3cd348873b729eb079cc26f8c2333a28247671bdb28",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/taxi_all.tsv",
        "taxi_all.tsv",
        "6d8ee9719b3d294b558eb53516c897108d1276e9dbcac0101d4e19a2ad801d20",
        zipped=False,
    ),
]


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt["datapath"], "msr_e2e")
    # define version if any
    version = "1.0"

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile
import gzip
import shutil

RESOURCES = [
    DownloadableFile(
        "https://dl.fbaipublicfiles.com/dpr/data/nq_gold_info/nq-train_gold_info.json.gz",
        'train_with_gold.json.gz',
        'b1a42d917a288ab330ee1ae4387efccbf6abba74d0da14e14a5ac38ff72d9ec8',
        zipped=False,
    ),
    DownloadableFile(
        "https://dl.fbaipublicfiles.com/dpr/data/nq_gold_info/nq-dev_gold_info.json.gz",
        'valid_with_gold.json.gz',
        '38e6ddeaf3918c99e2be920fd51aa587182a949ead4f8aec200f1999c8231df2',
        zipped=False,
    ),
    DownloadableFile(
        "https://dl.fbaipublicfiles.com/dpr/data/nq_gold_info/nq-test_gold_info.json.gz",
        'test_with_gold.json.gz',
        'a1588adca9cc1db68ae1cd1507b48de110e016740f77812b2e91d5bb4dc6cf17',
        zipped=False,
    ),
]

VERSION = 1


def build(opt):
    dpath = os.path.join(opt['datapath'], 'NaturalQuestionsOpen')
    version = str(VERSION)

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)
            if ".gz" in downloadable_file.file_name:
                with gzip.open(
                    os.path.join(dpath, downloadable_file.file_name), 'rb'
                ) as fin:
                    with open(
                        os.path.join(dpath, downloadable_file.file_name[:-3]), 'wb'
                    ) as fout:
                        shutil.copyfileobj(fin, fout)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Taskmaster-1 implementation for ParlAI.

Note that we have conversations structured both in the "TOD" format as well as those
from prior.
"""

from parlai.core.params import ParlaiParser
import os
import pandas as pd
from parlai.core.opt import Opt
import parlai.core.tod.tod_core as tod
from typing import Optional
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager

import parlai.tasks.taskmaster.build as build_
import parlai.core.tod.tod_agents as tod_agents

# Following is for legacy format
from parlai.core.teachers import FixedDialogTeacher
from . import tm_utils
import json


################### TOD Conversation format

SILENCE_TOKEN = "__SILENCE__"

# Faster to copy/paste this than parse a json file
ONTOLOGY = {
    "uber": {
        "id": "uber_lyft",
        "vertical": "ride_booking",
        "required": ["location.from", "location.to", "type.ride", "num.people"],
        "optional": [
            "price.estimate",
            "duration.estimate",
            "time.pickup",
            "time.dropoff",
        ],
    },
    "movie": {
        "id": "movie_ticket",
        "vertical": "ticket_booking",
        "required": [
            "name.movie",
            "name.theater",
            "num.tickets",
            "time.start",
            "location.theater",
            "price.ticket",
        ],
        "optional": ["type.screening", "time.end", "time.duration"],
    },
    "restaurant": {
        "id": "restaurant_reservation",
        "vertical": "reservation",
        "required": [
            "name.restaurant",
            "name.reservation",
            "num.guests",
            "time.reservation",
        ],
        "optional": ["type.seating", "location.restaurant"],
    },
    "coffee": {
        "id": "coffee_ordering",
        "vertical": "coffee_order",
        "required": ["location.store", "name.drink", "size.drink"],
        "optional": ["num.drink", "type.milk", "preference"],
    },
    "pizza": {
        "id": "pizza_ordering",
        "vertical": "pizza_order",
        "required": ["name.store", "name.pizza", "size.pizza"],
        "optional": ["type.topping", "type.crust", "preference", "location.store"],
    },
    "auto": {
        "id": "auto_repair",
        "vertical": "appointment",
        "required": ["name.store", "name.customer", "date.appt", "time.appt"],
        "optional": ["reason.appt", "name.vehicle", "year.vehicle", "location.store"],
    },
}


class Taskmaster1Parser(tod_agents.TodStructuredDataParser):
    """
    Abstract data loader.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "taskmaster-1")
        if shared is None:
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        chunks = []
        with PathManager.open(os.path.join(self.dpath, f"self-dialogs.json")) as f:
            subset = pd.read_json(f)
            chunks.append(subset)
        with PathManager.open(os.path.join(self.dpath, f"woz-dialogs.json")) as f:
            subset = pd.read_json(f)
            chunks.append(subset)
        chunks = pd.concat(chunks, axis=0)
        # deterministic shuffle data for splits
        chunks = chunks.sample(frac=1.0, random_state=42)
        split_size = len(chunks) // 10
        if fold == "train":
            chunks = chunks[: split_size * 8]
        elif fold == "valid":
            chunks = chunks[split_size * 8 : split_size * 9]
        elif fold == "test":
            chunks = chunks[split_size * 9 :]
        return chunks, ONTOLOGY

    def _parse_segment_to_slots(self, segment_list):
        result = {}
        for segment in segment_list:
            slot_name = segment["annotations"][0]["name"]
            slot_value = segment["text"]
            prefix_split_idx = slot_name.find(".")
            api_name = slot_name[:prefix_split_idx]
            slot_name = slot_name[prefix_split_idx + 1 :]
            result[slot_name] = slot_value
            result[tod.STANDARD_API_NAME_SLOT] = api_name
        return result

    def _get_utterance_and_slots_for_speaker(self, speaker, utterances, idx):
        utts = []
        slots = {}
        while idx < len(utterances):
            here = utterances[idx]
            if here["speaker"] != speaker:
                break
            utts.append(here["text"])
            slots.update(self._parse_segment_to_slots(here.get("segments", [])))
            idx += 1
        return idx, "\n".join(utts), slots

    def _parse_to_api_schema(self, raw):
        """
        NOTE: Format of ontology in this is different from TM2 + TM3. Need to figure out which is relevant for the domain.
        """
        result = {}
        for key, val in raw.items():
            here = {}
            here[tod.STANDARD_API_NAME_SLOT] = val["id"]
            here[tod.STANDARD_REQUIRED_KEY] = val.get("required", [])
            here[tod.STANDARD_OPTIONAL_KEY] = val.get("optional", [])
            result[key] = here
        return result

    def _get_turns_from_parsed(self, user_utt, api_calls, api_resps, sys_utt):
        result = [
            tod.TodStructuredRound(
                user_utt=user_utt,
                api_call_machine=api_calls,
                api_resp_machine=api_resps,
                sys_utt=sys_utt,
            )
        ]
        return result

    def setup_episodes(self, fold):
        """
        Parses into TodStructuredEpisode.
        """
        chunks, api_schema_raw = self._load_data(fold)
        api_schemas_machine = self._parse_to_api_schema(api_schema_raw)
        episodes = []
        for _, row in chunks.iterrows():
            utterances = row["utterances"][:]
            if not all(
                [
                    x.get("speaker") == "ASSISTANT" or x.get("speaker") == "USER"
                    for x in utterances
                ]
            ):
                # there's an example or two that causes things to infinite loop. >.>
                continue
            idx = 0
            rounds = []
            goal_calls = []
            if len(utterances) > 0 and utterances[0]["speaker"] == "ASSISTANT":
                (idx, sys_utt, _) = self._get_utterance_and_slots_for_speaker(
                    "ASSISTANT", utterances, idx
                )

                turns = self._get_turns_from_parsed(SILENCE_TOKEN, {}, {}, sys_utt)
                for t in turns:
                    rounds.append(t)

            while idx < len(utterances):
                (idx, user_utt, user_slots) = self._get_utterance_and_slots_for_speaker(
                    "USER", utterances, idx
                )
                (
                    idx,
                    sys_utt,
                    system_slots,
                ) = self._get_utterance_and_slots_for_speaker(
                    "ASSISTANT", utterances, idx
                )
                # The annotations in this dataset don't make sense as api responses but... we'll just roll.
                turns = self._get_turns_from_parsed(
                    user_utt, user_slots, system_slots, sys_utt
                )
                for t in turns:
                    rounds.append(t)
            apis = []
            for candidate_api in api_schemas_machine:
                if candidate_api in row["instruction_id"]:
                    apis.append(api_schemas_machine[candidate_api])
            episode = tod.TodStructuredEpisode(
                api_schemas_machine=apis,
                goal_calls_machine=goal_calls,
                rounds=rounds,
                delex=self.opt.get("delex", False),
            )
            episodes.append(episode)
        return episodes

    def get_id_task_prefix(self):
        return "Taskmaster1"

    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)


class SystemTeacher(Taskmaster1Parser, tod_agents.TodSystemTeacher):
    pass


class UserSimulatorTeacher(Taskmaster1Parser, tod_agents.TodUserSimulatorTeacher):
    pass


############ Legacy defined teachers


class SelfDialogueTeacher(FixedDialogTeacher):
    """
    Teacher for written two-person dialogues with labels being responses for the
    previous statement.

    The data is traversed twice (doubled), once for modelling USER replies and once for
    modelling ASSISTANT replies.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt)
        opt['fn'] = "self-dialogs.json"

        if shared and 'convos' in shared:
            # another instance was set up already, just reference its data
            self.convos = shared['convos']
            self.ep_cheat_sheet = shared['ep_cheat_sheet']
            self.num_ex = shared['num_ex']
        else:
            # need to set up data from scratch
            self.ep_cheat_sheet = {}  # Stores imp. info. about each episode
            data_path = tm_utils._path(opt)
            self.num_ex = 0
            self._setup_data(data_path, opt)

        self.reset()

    def _setup_data(self, data_path, opt):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.convos = json.load(data_file)
        # Pre-processing
        convos_update = []
        for convo in self.convos:
            conversation = convo['utterances']
            # Filter out single greet messages
            if len(conversation) > 1:
                self.ep_cheat_sheet[
                    len(self.ep_cheat_sheet)
                ] = tm_utils.gen_ep_cheatsheet(conversation)
                curr_cheatsheet = self.ep_cheat_sheet[len(self.ep_cheat_sheet) - 1]
                self.num_ex += (
                    curr_cheatsheet[tm_utils.USER_NUM_EX]
                    + curr_cheatsheet[tm_utils.ASSIS_NUM_EX]
                )
                convos_update += [conversation]
        self.convos = convos_update

    def num_examples(self):
        return self.num_ex

    def num_episodes(self):
        # For two passes over the data: Once to teach USER and once to teach ASSISTANT
        return len(self.convos) * 2

    def get(self, episode_idx, entry_idx):
        conversation = self.convos[episode_idx % len(self.convos)]
        if episode_idx < len(self.convos):
            # USER then ASSISTANT mode [First pass]
            ep_done = (
                entry_idx * 2
                == self.ep_cheat_sheet[episode_idx][tm_utils.LAST_USER_IDX]
            )
            predecessor = conversation[entry_idx * 2]['text']
            successor = conversation[entry_idx * 2 + 1]['text']
        else:
            # ASSISTANT then USER mode [Second pass]
            ep_done = (
                entry_idx * 2 + 1
                == self.ep_cheat_sheet[episode_idx % len(self.convos)][
                    tm_utils.LAST_ASSISTANT_IDX
                ]
            )
            predecessor = conversation[entry_idx * 2 + 1]['text']
            successor = conversation[entry_idx * 2 + 2]['text']

        action = {
            'id': self.id,
            'text': predecessor,
            'episode_done': ep_done,
            'labels': [successor],
        }

        return action


class WozDialogueTeacher(FixedDialogTeacher):
    """
    Teacher for spoken two-person dialogues with labels being responses for the previous
    statement.

    The data is traversed twice (doubled), once for modelling USER replies and once for
    modelling ASSISTANT replies.
    """

    def __init__(self, opt, shared=None):
        opt['fn'] = "woz-dialogs.json"
        super().__init__(opt)

        if shared and 'convos' in shared:
            # another instance was set up already, just reference its data
            self.convos = shared['convos']
            self.episode_map = shared['episode_map']
            self.ep_cheat_sheet = shared['ep_cheat_sheet']
            self.num_ex = shared['num_ex']
        else:
            # need to set up data from scratch
            self.ep_cheat_sheet = {}  # Stores imp. info. about each episode

            # Not all episodes have relevant examples for both USER and ASSISTANT
            # episode_map keeps track of which episode index is useful for which speaker
            # Need to do this otherwise might end up with a situation where we cannot
            # return anything in action
            self.episode_map = {}
            self.episode_map["U"] = {}
            self.episode_map["A"] = {}
            self.num_ex = 0
            data_path = tm_utils._path(opt)
            self._setup_data(data_path, opt)

        self.reset()

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Corrupt-Example-Arguments')
        agent.add_argument(
            '--exclude-invalid-data',
            type='bool',
            default=True,
            help='Whether to include corrupt examples in the data',
        )
        return parser

    def _setup_data(self, data_path, opt):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.convos = json.load(data_file)
        # Pre-processing
        convos_update = []
        for convo in self.convos:
            conversation, corrupted = tm_utils.smoothen_convo(convo, opt)
            # Filter out single greet messages and corrupted examples
            if len(conversation) > 1 and not corrupted:
                actual_ep_idx = len(self.ep_cheat_sheet)
                self.ep_cheat_sheet[actual_ep_idx] = tm_utils.gen_ep_cheatsheet(
                    conversation
                )
                curr_cheatsheet = self.ep_cheat_sheet[len(self.ep_cheat_sheet) - 1]
                # calc number of examples (done here to prevent double counting if done later)
                self.num_ex += (
                    curr_cheatsheet[tm_utils.USER_NUM_EX]
                    + curr_cheatsheet[tm_utils.ASSIS_NUM_EX]
                )
                # User example exists
                if curr_cheatsheet[tm_utils.USER_NUM_EX] != 0:
                    u_idx = len(self.episode_map["U"])
                    self.episode_map["U"][u_idx] = actual_ep_idx
                # Assistant example exists
                if curr_cheatsheet[tm_utils.ASSIS_NUM_EX] != 0:
                    a_idx = len(self.episode_map["A"])
                    self.episode_map["A"][a_idx] = actual_ep_idx
                convos_update += [conversation]
        self.convos = convos_update

    def num_examples(self):
        return self.num_ex

    def num_episodes(self):
        # For two passes over the data: Once to teach USER and once to teach ASSISTANT
        return len(self.episode_map["U"]) + len(self.episode_map["A"])

    def get(self, episode_idx, entry_idx):
        if episode_idx < len(self.episode_map["U"]):
            # USER then ASSISTANT mode [First pass]
            true_idx = self.episode_map["U"][episode_idx]
            conversation = self.convos[true_idx]
            convo_cheat_sheet = self.ep_cheat_sheet[true_idx]
            first_entry_idx, last_entry_idx = (
                convo_cheat_sheet[tm_utils.FIRST_USER_IDX],
                convo_cheat_sheet[tm_utils.LAST_USER_IDX],
            )
        else:
            # ASSISTANT then USER mode [Second pass]
            episode_idx -= len(
                self.episode_map["U"]
            )  # Didn't use '%' because the two maybe unequal in length
            true_idx = self.episode_map["A"][episode_idx]
            conversation = self.convos[true_idx]
            convo_cheat_sheet = self.ep_cheat_sheet[true_idx]
            first_entry_idx, last_entry_idx = (
                convo_cheat_sheet[tm_utils.FIRST_ASSISTANT_IDX],
                convo_cheat_sheet[tm_utils.LAST_ASSISTANT_IDX],
            )

        starts_at_odd = first_entry_idx % 2 != 0
        if starts_at_odd:
            predecessor = conversation[entry_idx * 2 + 1]['text']
            successor = conversation[entry_idx * 2 + 2]['text']
            ep_done = entry_idx * 2 + 1 == last_entry_idx
        else:
            predecessor = conversation[entry_idx * 2]['text']
            successor = conversation[entry_idx * 2 + 1]['text']
            ep_done = entry_idx * 2 == last_entry_idx

        action = {
            'id': self.id,
            'text': predecessor,
            'episode_done': ep_done,
            'labels': [successor],
        }

        return action


class SelfDialogueSegmentTeacher(FixedDialogTeacher):
    """
    Teacher for written two-person dialogues with labels being relevant/useful parts in
    the input sentence.

    The different datatypes of the labels within the data have also been encoded as
    `label_types`
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt)
        opt['fn'] = "self-dialogs.json"

        if shared and 'convos' in shared:
            # another instance was set up already, just reference its data
            self.convos = shared['convos']
            self.num_ex = shared['num_ex']
        else:
            # need to set up data from scratch
            data_path = tm_utils._path(opt)
            self.num_ex = 0
            self._setup_data(data_path, opt)
        self.reset()

    def get(self, episode_idx, entry_idx):
        conversation = self.convos[episode_idx]
        conv_len = len(conversation['utterances'])
        utterance = conversation['utterances'][entry_idx]['text']

        # Check if episode is complete
        ep_done = entry_idx == conv_len - 1
        action = {'id': self.id, 'text': utterance, 'episode_done': ep_done}

        # Setup Labels as "text" from segments
        action['labels'] = []
        action['label_types'] = []
        segments = conversation['utterances'][entry_idx]["segments"]
        for segment in segments:
            action['labels'] += [segment["text"]]
            tmp = []
            for annot in segment["annotations"]:
                tmp += [annot["name"]]
            action['label_types'] += [tmp]

        return action

    def num_examples(self):
        return self.num_ex

    def num_episodes(self):
        return len(self.convos)

    def _setup_data(self, data_path, opt):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.convos = json.load(data_file)

        # Filter out instances which do not have "segment" in them
        convos_updated = []
        for convo in self.convos:
            updated_dialog = []
            for i in range(len(convo['utterances'])):
                if "segments" in convo['utterances'][i]:
                    updated_dialog += [convo['utterances'][i]]
            convo['utterances'] = updated_dialog
            if convo['utterances']:
                convos_updated += [convo]
                self.num_ex += len(convo['utterances'])
        self.convos = convos_updated


class DefaultTeacher(SelfDialogueTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base script for running TOD model-model chats.
"""
from parlai.core.metrics import InterDistinctMetric

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script

import glob
import json


@register_script("get_interdistinct_on_conversations")
class GetInterdistinctOnConversationsScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(False, False)
        parser.add_argument("-p", "--path", required=True, type=str)
        return parser

    def run(self):
        opt = self.opt
        paths = glob.glob(f"{opt['path']}/*_conversations.jsonl")

        for path in paths:
            name = path.split("/")[-1]
            sys_utt_one = InterDistinctMetric.compute("", 1)
            sys_utt_two = InterDistinctMetric.compute("", 2)
            user_utt_one = InterDistinctMetric.compute("", 1)
            user_utt_two = InterDistinctMetric.compute("", 2)

            with open(path) as f:
                for line_raw in f:
                    line = json.loads(line_raw)["dialog"]
                    for turn in line:
                        if len(turn) < 4:
                            continue
                        user_utt = turn[0]["text"]
                        if user_utt.startswith("APIS: "):
                            continue
                        user_utt_one += InterDistinctMetric.compute(user_utt, 1)
                        user_utt_two += InterDistinctMetric.compute(user_utt, 2)

                        sys_utt = turn[3]["text"]
                        sys_utt_one += InterDistinctMetric.compute(sys_utt, 1)
                        sys_utt_two += InterDistinctMetric.compute(sys_utt, 2)

            print(
                ",".join(
                    [
                        str(x)
                        for x in [
                            name,
                            user_utt_one.value(),
                            user_utt_two.value(),
                            sys_utt_one.value(),
                            sys_utt_two.value(),
                        ]
                    ]
                )
            )


if __name__ == "__main__":
    GetInterdistinctOnConversationsScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
MSR-E2E implementation for ParlAI.

No official train/valid/test splits are available of public data, so we make our own.

We assume inform slots from the agent are API responses and request/inform slots from the user is an API call. It's not quite how things are supposed to work, but the `dialogue act` setup is not super well standardized within the dataset.
"""

from parlai.core.params import ParlaiParser
import copy
import os
import csv
from collections import Counter
from parlai.core.opt import Opt
import parlai.core.tod.tod_core as tod
from parlai.utils.misc import warn_once
from typing import Optional
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager

import parlai.tasks.msr_e2e.build as build_
import parlai.core.tod.tod_agents as tod_agents


DOMAINS = ["movie", "restaurant", "taxi"]

# Just going to copy/paste these since it's faster than parsing 3 separate files
# They are in `system/src/deep_dialog/data_<domain>/<domain>_slots.txt` in the original data
SLOT_NAMES = {
    "movie": [
        "actor",
        "actress",
        "city",
        "closing",
        "critic_rating",
        "date",
        "schema",
        "distanceconstraints",
        "genre",
        "greeting",
        "implicit_value",
        "movie_series",
        "moviename",
        "mpaa_rating",
        "numberofpeople",
        "numberofkids",
        "taskcomplete",
        "other",
        "price",
        "seating",
        "starttime",
        "state",
        "theater",
        "theater_chain",
        "video_format",
        "zip",
        "result",
        "ticket",
        "mc_list",
    ],
    "restaurant": [
        "address",
        "atmosphere",
        "choice",
        "city",
        "closing",
        "cuisine",
        "date",
        "distanceconstraints",
        "dress_code",
        "food",
        "greeting",
        "implicit_value",
        "mealtype",
        "numberofpeople",
        "numberofkids",
        "occasion",
        "other",
        "personfullname",
        "phonenumber",
        "pricing",
        "rating",
        "restaurantname",
        "restauranttype",
        "seating",
        "starttime",
        "state",
        "zip",
        "result",
        "mc_list",
        "taskcomplete",
        "reservation",
    ],
    "taxi": [
        "car_type",
        "city",
        "closing",
        "date",
        "distanceconstraints",
        "dropoff_location",
        "greeting",
        "name",
        "numberofpeople",
        "other",
        "pickup_location",
        "dropoff_location_city",
        "pickup_location_city",
        "pickup_time",
        "state",
        "cost",
        "taxi_company",
        "mc_list",
        "taskcomplete",
        "taxi",
        "zip",
        "result",
        "mc_list",
    ],
}

SLOT_NAMES = {
    k: [{"api_name": k, "optArg": " | ".join(v)}] for k, v in SLOT_NAMES.items()
}


class MsrE2EParser(tod_agents.TodStructuredDataParser):
    """
    Abstract data loader.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            "--msre2e-domains",
            nargs="+",
            default=DOMAINS,
            choices=DOMAINS,
            help="Uses last passed in configuration.",
        )
        parser.add_argument(
            "--use-cumulative-api-calls",
            type=bool,
            default=True,
            help="Have API Call/API response turns only when an API response"
            "slot exist. Accumulate all API call slots with same API call name",
        )
        return super().add_cmdline_args(parser, partial_opt)

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "msr_e2e")
        if shared is None:
            warn_once("MsrE2E is a beta dataset, and format may significantly change.")
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold, domains):
        chunks = []
        for section in domains:
            domain = []
            with PathManager.open(os.path.join(self.dpath, section + "_all.tsv")) as f:
                reader = csv.reader(f, delimiter="\t")
                next(reader)
                lines = list(reader)
            episode = []
            prev_idx = 0
            for line in lines:
                data = {}
                data["id"] = line[0]
                data["speaker"] = line[3]
                data["text"] = line[4]
                data["dialogue_acts"] = line[5:]
                data["domain"] = section
                if prev_idx != data["id"]:
                    domain.append(episode)
                    episode = []
                    prev_idx = data["id"]
                episode.append(data)
            domain.append(episode)
            chunks.append(domain)
        # deterministic shuffle data for splits
        return DatatypeHelper.split_subset_data_by_fold(fold, chunks, 0.8, 0.1, 0.1)

    def _parse_dialogue_act(self, act, domain):
        act = (
            act.replace("inform(", "")
            .replace("request(", "")
            .replace("multiple_choice(", "")
        )
        act = act[:-1]

        args = act.split(";")
        result = {}
        for arg in args:
            params = arg.split("=")
            key = params[0]
            if (
                key == "other"
            ):  # This is all stuff the E2E model should be able to pick up on its own.
                continue
            if (
                len(params) == 1
            ):  # MSR_E2E has this as a "what explicit information do we want" slot, but it's not super consistent
                continue
            result[key] = "=".join(params[1:])
        if len(result) > 0:
            result[tod.STANDARD_API_NAME_SLOT] = domain
        return result

    def _get_utterance_and_api_call_for_speaker(self, speaker, utterances, idx):
        utts = []
        slots = {}
        while idx < len(utterances):
            here = utterances[idx]
            if here["speaker"] != speaker:
                break
            utts.append(here["text"])
            for act in utterances[idx]["dialogue_acts"]:
                if speaker == "agent" and not (
                    act.startswith("inform") or act.startswith("multiple_choice")
                ):
                    continue
                if speaker == "user" and not (
                    act.startswith("inform") or act.startswith("request")
                ):
                    continue
                slots.update(self._parse_dialogue_act(act, utterances[0]["domain"]))
            idx += 1
        return idx, "\n".join(utts), slots

    def setup_episodes(self, fold):
        """
        Parses into TodStructuredEpisode.
        """
        domains = self.opt.get("msre2e_domains", DOMAINS)
        chunks = self._load_data(fold, domains)
        domains_cnt = Counter()
        episodes = []
        for utterances in chunks:
            if len(utterances) < 1:
                continue
            domain = utterances[0]["domain"]
            domains_cnt[domain] += 1
            idx = 0
            rounds = []
            goal_calls = []
            if len(utterances) > 0 and utterances[0]["speaker"] == "agent":
                idx, sys_utt, api_resp = self._get_utterance_and_api_call_for_speaker(
                    "agent", utterances, idx
                )
                r = tod.TodStructuredRound(
                    user_utt=tod.CONST_SILENCE,
                    api_resp_machine=api_resp,
                    sys_utt=sys_utt,
                )
                rounds.append(r)

            cum_api_call = {}
            while idx < len(utterances):
                idx, user_utt, api_call = self._get_utterance_and_api_call_for_speaker(
                    "user", utterances, idx
                )
                idx, sys_utt, api_resp = self._get_utterance_and_api_call_for_speaker(
                    "agent", utterances, idx
                )
                if not self.opt["use_cumulative_api_calls"]:
                    r = tod.TodStructuredRound(
                        user_utt=user_utt,
                        api_call_machine=api_call,
                        api_resp_machine=api_resp,
                        sys_utt=sys_utt,
                    )
                else:
                    cum_api_call.update(api_call)
                    r = tod.TodStructuredRound(
                        user_utt=user_utt,
                        api_call_machine=copy.deepcopy(cum_api_call)
                        if len(api_resp) > 0
                        else {},
                        api_resp_machine=api_resp if len(api_resp) > 0 else {},
                        sys_utt=sys_utt,
                    )

                rounds.append(r)
                if len(api_call) > 0:
                    goal_calls.append(api_call)

            episode = tod.TodStructuredEpisode(
                domain=domain,
                api_schemas_machine=SLOT_NAMES[domain],
                goal_calls_machine=goal_calls,
                rounds=rounds,
                delex=self.opt.get("delex", False),
            )
            episodes.append(episode)
        return episodes

    def get_id_task_prefix(self):
        return "MsrE2E"

    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)


class SystemTeacher(MsrE2EParser, tod_agents.TodSystemTeacher):
    pass


class UserSimulatorTeacher(MsrE2EParser, tod_agents.TodUserSimulatorTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.opt import Opt
import parlai.utils.logging as logging
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.msc.agents as msc

from projects.bb3.tasks.r2c2_dialogue_tasks import (
    WoiSearchDialogueTeacher as woi_srm,
    WowSearchDialogueTeacher as wow_srm,
    MsMarcoSearchDialogueTeacher as msmarco_srm,
    FitsSearchDialogueTeacher as fits_srm,
    FunpediaSearchDialogueTeacher as funpedia_srm,
    GoogleSgdSearchDialogueTeacher as googlesgd_srm,
    TaskmasterSearchDialogueTeacher as taskmaster_srm,
    Taskmaster2SearchDialogueTeacher as taskmaster2_srm,
    Taskmaster3SearchDialogueTeacher as taskmaster3_srm,
    Convai2EntityDialogueTeacher as convai2_crm,
    EDEntityDialogueTeacher as ed_crm,
    BSTEntityDialogueTeacher as bst_crm,
    MSCEntityDialogueTeacher as msc_crm,
    SaferdialoguesVanillaDialogueTeacher as Saferdialogues_vrm,
    WoiVanillaDialogueTeacher as woi_vrm,
    WowVanillaDialogueTeacher as wow_vrm,
    LightVanillaDialogueTeacher as light_vrm,
    LightWildVanillaDialogueTeacher as lightwild_vrm,
    EDVanillaDialogueTeacher as ed_vrm,
    Convai2VanillaDialogueTeacher as convai2_vrm,
    MSCVanillaDialogueTeacher as msc_vrm,
    Convai2StyleGroundingDialogueTeacher as convai2_grm,
    BSTStyleGroundingDialogueTeacher as bst_grm,
    get_memory_knowledge_pers_overlap_task_mutator,
    get_memory_knowledge_utt_overlap_task_mutator,
    get_memory_dialogue_task_mutator,
)

from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin


def get_prefix_speakers_opt_mutators(opt: Opt):
    mutators = '+'.join(['prefix_speakers_opt'])
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


def get_dialogue_opt_mutators(opt: Opt):
    mutators = '+'.join(
        ['prefix_speakers_opt', 'format_response_tasks_for_decoder_only']
    )
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


def get_vanilla_dialogue_opt_mutators(opt: Opt):
    mutators = '+'.join(
        ['prefix_speakers_opt', 'format_vanilla_dialogue_for_decoder_only']
    )
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


def get_light_dialogue_opt_mutators(opt: Opt):
    mutators = '+'.join(['prefix_speakers_opt', 'format_light_tasks_for_decoder_only'])
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


def get_style_dialogue_opt_mutators(opt: Opt):
    mutators = '+'.join(
        ['prefix_speakers_opt', 'format_style_grounding_tasks_for_decoder_only']
    )
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2EntityDialogueTeacher(convai2_crm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class EDEntityDialogueTeacher(ed_crm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCEntityDialogueTeacher(msc_crm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTEntityDialogueTeacher(bst_crm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2MemoryDialogueFromPersOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


class Convai2MemoryDialogueFromUttOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


class MsMarcoSearchDialogueTeacher(msmarco_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class WowSearchDialogueTeacher(wow_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class WoiSearchDialogueTeacher(woi_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class FitsSearchDialogueTeacher(fits_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class FunpediaSearchDialogueTeacher(funpedia_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class GoogleSgdSearchDialogueTeacher(googlesgd_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class TaskmasterSearchDialogueTeacher(taskmaster_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class Taskmaster2SearchDialogueTeacher(taskmaster2_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class Taskmaster3SearchDialogueTeacher(taskmaster3_srm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class WowVanillaDialogueTeacher(wow_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_prefix_speakers_opt_mutators(opt)
        super().__init__(opt, shared)


class WoiVanillaDialogueTeacher(woi_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_prefix_speakers_opt_mutators(opt)
        super().__init__(opt, shared)


class EDVanillaDialogueTeacher(ed_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_prefix_speakers_opt_mutators(opt)
        super().__init__(opt, shared)


class SaferdialoguesVanillaDialogueTeacher(Saferdialogues_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_prefix_speakers_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2VanillaDialogueTeacher(convai2_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_vanilla_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCVanillaDialogueTeacher(msc_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_vanilla_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class LightVanillaDialogueTeacher(light_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_light_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class LightWildVanillaDialogueTeacher(lightwild_vrm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_light_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2StyleGroundingDialogueTeacher(convai2_grm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_style_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTStyleGroundingDialogueTeacher(bst_grm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_style_dialogue_opt_mutators(opt)
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile
import json
from parlai.utils.io import PathManager

RESOURCES = [
    ########En########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/En_persona_test.json',
        'En_test_tmp.json',
        '8baa09a8064a22967544f501821aa114393a59339c0559da8afa160966ba87c9',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/En_persona_train.json',
        'En_train_tmp.json',
        'e23112bba7320f798b07afb4c5acc3edad2a2ccb7df5cc46f141a0c79ff4665c',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/En_persona_valid.json',
        'En_valid_tmp.json',
        '08ed3d41c5b0681c2d125a5312b43d926a8a5aa1d10a5df655d17f4c56dab635',
        zipped=False,
    ),
    #######Fr########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Fr_persona_train_corrected.json',
        'Fr_train_tmp.json',
        '40e66e91aa6360eeda642c2c674f03c52854626eaf35da50084d26ff42f61292',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Fr_persona_split_test_human_annotated.json',
        'Fr_test_tmp.json',
        '0783fcf01bdf4c27ec28120a9b23bf4f0248e97ae476e03cd7e759cb2667ab23',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Fr_persona_split_valid_human_annotated.json',
        'Fr_valid_tmp.json',
        '8ad86b05aabfadedba7863828b1cc4fdff0926ebf476121268089ac7ed9af149',
        zipped=False,
    ),
    ########Id########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Id_persona_train_corrected.json',
        'Id_train_tmp.json',
        'fee1a9769fe707fd09401c33bdf3b3cd4f7b5fd100998577f3f179c42423bc4f',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Id_persona_split_test_human_annotated.json',
        'Id_test_tmp.json',
        'e0bcd3c02f318f4381c42798a2ce6e0a10237b6998793076fbf97f633a9f2563',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Id_persona_split_valid_human_annotated.json',
        'Id_valid_tmp.json',
        '8f70cab662f082ae3ee2abce9e9cac619ebc632a2820d93414a59005dbf75d7e',
        zipped=False,
    ),
    ########It########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/It_persona_train_corrected.json',
        'It_train_tmp.json',
        '9636893050ad16dc2daabfe8bde6979c9c780c5da9d0790f5668e198aac18b8f',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/It_persona_split_test_human_annotated.json',
        'It_test_tmp.json',
        '690103031791fd5c6763176a074cdb65e249adc784e2d577110c2d9430a02a87',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/It_persona_split_valid_human_annotated.json',
        'It_valid_tmp.json',
        '720dc91d3f9bc6a56ac229c6800cde63c71677a3db8d0ace7a59a7f94d89df3d',
        zipped=False,
    ),
    ########Jp########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Jp_persona_train_corrected.json',
        'Jp_train_tmp.json',
        '808ee29c79303e1300b38aceee79111ffe1fd2a2facb09a0956615a61d840738',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Jp_persona_split_test_human_annotated.json',
        'Jp_test_tmp.json',
        '682fed16148517097437942088d225bd728cb7b41aa390559681ae73e5e6848f',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Jp_persona_split_valid_human_annotated.json',
        'Jp_valid_tmp.json',
        'a86bb811364d100bc77ddc8038265df1e01bbc3be095e56c6ec8f179e6365d75',
        zipped=False,
    ),
    ########Ko########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Ko_persona_train_corrected.json',
        'Ko_train_tmp.json',
        '105d6a08d02e76f1d006edb1819b96a8b5fa8d94b3ed278936bcf171368809b7',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Ko_persona_split_test_human_annotated.json',
        'Ko_test_tmp.json',
        'f7ac6bd2aec7014a28d34bb34dceed653b389ce25d21ca770e77142578dc70a6',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Ko_persona_split_valid_human_annotated.json',
        'Ko_valid_tmp.json',
        '188470f863f639946bc8248a9f6aa1e589b41ec61792b88b81e7a95c72deeae0',
        zipped=False,
    ),
    ########Zh########
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Zh_persona_train_corrected.json',
        'Zh_train_tmp.json',
        'e07899fa91edd127ec77502bd604693c40e264b60225976e2ac6ed145d080323',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Zh_persona_split_test_human_annotated.json',
        'Zh_test_tmp.json',
        '0767a4a27c765277792597502f57ea8bb80bf7be94613b0d833107f66a7d3512',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/HLTCHKUST/Xpersona/master/dataset/Zh_persona_split_valid_human_annotated.json',
        'Zh_valid_tmp.json',
        'cfa90117d73fe294a1b776b2d1c7b53711bfcd724f5833204726c910dad5482d',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'XPersona')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)
        _create_parlai_format(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)


def _create_parlai_format(dpath: str):
    datatypes = ['train', 'valid', 'test']
    languages = ['En_', 'Zh_', 'Fr_', 'Ko_', 'Id_', 'Jp_', 'It_']
    for language in languages:
        for datatype in datatypes:
            datatype_full = language + datatype + '_tmp'
            datatype_rename = language + datatype
            load_path = os.path.join(dpath, f'{datatype_full}.json')
            save_path = os.path.join(dpath, f'{datatype_rename}.txt')
            with PathManager.open(load_path, 'r', encoding='utf8') as f_read:
                data = json.load(f_read)
            with PathManager.open(save_path, 'w', encoding='utf8') as f_write:
                for content in data:
                    line_num = 0
                    personas = content['persona']
                    dialogs = content['dialogue']
                    for persona in personas:
                        line_num += 1
                        f_write.write(str(line_num) + ' your persona:' + persona + '\n')
                    for utterance_A, utterance_B in dialogs:
                        line_num += 1
                        f_write.write(f"{line_num} {utterance_A}\t{utterance_B}\n")
            os.remove(load_path)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Automated Expanded Attention + Multi-Objective.

Combines both expanded attention and multi-objective training objective
"""
from parlai.zoo.light_whoami.whoami_download import download_with_model_type


def download(datapath):
    download_with_model_type(datapath, 'expanded_and_multiobjective_1024', 'v1.0')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from git import Repo
import os

import parlai.core.build_data as build_data
from parlai.core.opt import Opt
import parlai.utils.logging as logging


GH_REPO = 'https://github.com/JerichoWorld/JerichoWorld.git'

DATASET_NAME = 'jericho'
VERSION = '1.0'


def build(opt: Opt) -> None:
    dpath = os.path.join(opt['datapath'], DATASET_NAME)
    if build_data.built(dpath, VERSION):
        logging.debug('Data was already built. Skipping the data building.')
        return

    if os.path.exists(dpath):
        logging.debug(f'Removing old/corrupted data in {dpath}.')
        build_data.remove_dir(dpath)

    logging.info(
        f'[building data: {dpath}]\nThis may take a while but only heppens once.'
    )
    logging.info(f'Cloning Github repo {GH_REPO}')
    temp_path = os.path.join(dpath, "temp")
    Repo.clone_from(GH_REPO, temp_path)
    build_data.untar(temp_path, 'data.zip')

    # Copying the unzipped data files to the dpath
    for dt in ('train', 'test'):
        fname = f'{dt}.json'
        fsource = os.path.join(temp_path, 'data', fname)
        fdest = os.path.join(dpath, fname)
        os.rename(fsource, fdest)

    # Removing unused files from the repository
    build_data.remove_dir(temp_path)

    build_data.mark_done(dpath, VERSION)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base script for running TOD model-model chats.
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript
import glob
import json
import os


class GetQuickEvalStatsDirScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(False, False)
        parser.add_argument("-p", "--path", required=True, type=str)
        return parser

    def run(self):
        opt = self.opt
        path = opt["path"]

        stuff = {}

        for eval_stats in glob.glob(f"{path}/*/eval_stats.json"):
            with open(eval_stats) as f:
                eval_data = json.load(f)

            base_path = os.path.abspath(eval_stats).replace("eval_stats.json", "/")

            base = ""
            multitask = ""
            if not os.path.isfile(base_path + "run.sh"):
                continue
            with open(base_path + "run.sh") as run_file:
                for line in run_file:
                    if "zoo:bart" in line:
                        base = "BartOnly"

            with open(base_path + "model.opt") as opt_file:
                print(base_path)
                opt = json.load(opt_file)
                lr = opt["learningrate"]
                multitask = opt.get("multitask_weights", "")
                yes_api = opt.get("api_schemas", False)
            if multitask != "":
                multitask = "_" + "".join([str(x) for x in multitask])

            orig_path = "/".join(os.path.abspath(base_path).split("/")[-3:])

            ppl = eval_data["report"].get("ppl", "")
            token_em = eval_data["report"].get("token_em", "")
            jga = eval_data["report"].get("jga", "")
            jga_n = eval_data["report"].get("jga_noempty", "")
            jga_e = eval_data["report"].get("jga_empty", "")

            if yes_api:  # hack cause I'm annoyed at things being in wrong alpha order
                yes_api = "True"
            else:
                yes_api = "false"

            maybe_nshot = ""
            root = base_path
            eval_file = glob.glob(f"{root}/*mm_eval*.json")
            if len(eval_file) > 0 and (
                "percent" in eval_file[0] or "nshot" in eval_file[0]
            ):
                maybe_nshot = eval_file[0][len(root + "mm_eval_") : -len(".json")]

            base += f"{len(opt.get('task').split(','))}-{yes_api}-APIS__{maybe_nshot}{base}_{lr}{multitask}"
            metrics = [orig_path, base, ppl, token_em, jga, jga_n, jga_e]

            if os.path.isfile(eval_stats.replace("eval_stats", "user_eval_stats")):
                with open(eval_stats.replace("eval_stats", "user_eval_stats")) as f:
                    metrics.append(json.load(f)["report"]["ppl"])
            else:
                metrics.append("DNE")

            metrics.append(" ")

            root = base_path
            maybe_mm_stats = glob.glob(f"{root}/mm_eval")
            if len(maybe_mm_stats) == 0:
                maybe_mm_stats = glob.glob(f"{root}/*mm_eval*.json")
            if len(maybe_mm_stats) == 0:
                maybe_mm_stats = glob.glob(f"{root}/*Mult*.json")
            if len(maybe_mm_stats) > 0:
                with open(maybe_mm_stats[0]) as f:
                    report = json.load(f)["report"]
                    if "synthetic_task_success" in report:
                        tsr = report["synthetic_task_success"]
                    elif "all_goals_hit" in report:
                        tsr = report["all_goals_hit"]
                    else:
                        tsr = "DNE"
                    metrics.append(tsr)
            else:
                metrics.append("DNE")

            metrics.append(" ")

            if os.path.isfile(eval_stats.replace("eval_stats", "in_eval_stats")):
                with open(eval_stats.replace("eval_stats", "in_eval_stats")) as f:
                    blob = json.load(f)["report"]
                    metrics.append(blob.get("ppl", ""))
                    metrics.append(blob.get("jga", ""))
                    metrics.append(blob.get("jga_noempty", ""))
                    metrics.append(blob.get("jga_empty", ""))
            else:
                metrics.append("DNE")
                metrics.append("DNE")
                metrics.append("DNE")
                metrics.append("DNE")

            if os.path.isfile(eval_stats.replace("eval_stats", "in_user_eval_stats")):
                with open(eval_stats.replace("eval_stats", "in_user_eval_stats")) as f:
                    metrics.append(json.load(f)["report"]["ppl"])
            else:
                metrics.append("DNE")

            if os.path.isfile(eval_stats.replace("eval_stats", "test_eval_stats")):
                with open(eval_stats.replace("eval_stats", "test_eval_stats")) as f:
                    blob = json.load(f)["report"]
                    metrics.append(blob.get("ppl", ""))
                    metrics.append(blob.get("jga", ""))
                    metrics.append(blob.get("jga_noempty", ""))
                    metrics.append(blob.get("jga_empty", ""))
            else:
                metrics.append("DNE")
                metrics.append("DNE")
                metrics.append("DNE")
                metrics.append("DNE")

            if os.path.isfile(eval_stats.replace("eval_stats", "test_user_eval_stats")):
                with open(
                    eval_stats.replace("eval_stats", "test_user_eval_stats")
                ) as f:
                    metrics.append(json.load(f)["report"]["ppl"])
            else:
                metrics.append("DNE")

            stuff[base] = ",".join([str(x) for x in metrics])

        ordering = [
            "orig_path",
            "base",
            "sys_ppl",
            "token_em",
            "jga",
            "jga_noempty",
            "jga_empty",
            "user_ppl",
            "space",
            "sTsr",
            "space",
            "in_sys_ppl",
            "in_jga",
            "in_jga_noempty",
            "in_jga_empty",
            "in_user_ppl",
            "test_sys_ppl",
            "test_jga",
            "test_jga_noempty",
            "test_jga_empty",
            "test_user_ppl",
        ]
        print(",".join(ordering))
        for key in sorted(stuff.keys()):
            print(stuff[key])


if __name__ == "__main__":
    GetQuickEvalStatsDirScript.main()
